# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import AxisItem as PgAxisItem, DateAxisItem
from qtpy.QtCore import Qt, Signal

from karabogui.fonts import get_qfont
from karabogui.graph.common.const import (
    AXIS_ITEMS, AXIS_X, AXIS_Y, X_AXIS_HEIGHT, Y_AXIS_WIDTH, get_alarm_string,
    get_state_string)
from karabogui.graph.common.enums import AxisType


def get_axis_font(size):
    """Return an appropriate karabo font with a point size `size`"""
    font = f"Source Sans Pro,{size},-1,5,50,0,0,0,0,0"
    return get_qfont(font)


class AxisItem(PgAxisItem):
    axisDoubleClicked = Signal()

    axisStyle = {
        "autoExpandTextSpace": False,
        "tickTextWidth": 50,
        "tickTextHeight": 24,
        "tickFont": get_axis_font(size=10)}

    def __init__(self, orientation, showValues=True):
        """Base class for pretty axis items.

        :param orientation: Location of the axis item in the plot
            Choose from ["left", "right", "top", "bottom"]
        :param showValues: Set if it will show ticks and labels.
            Usually major axis has the ticks (bottom and left for normal plots)
        """
        super().__init__(orientation, showValues=showValues)
        self.enableAutoSIPrefix(False)
        if showValues:
            self.setStyle(**self.axisStyle)

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.axisDoubleClicked.emit()
            event.accept()
            return
        super().mouseDoubleClickEvent(event)

    def setLabel(self, text=None, units=None, unitPrefix=None, **args):
        """Reimplemented because we don't want labels for minor axes"""
        if not self.style["showValues"]:
            return
        super().setLabel(text=text, units=units, unitPrefix=unitPrefix, **args)

    def showLabel(self, show=True):
        """Always do not show label if it is not a tick axis"""
        if not self.style["showValues"]:
            show = False
        super().showLabel(show)

    def mouseDragEvent(self, event):
        """Reimplemented function of PyQt

        PyQtGraph axis items have drag events which can enables panning
        of the viewbox.
        This is a problem for axis with grids because it is resized to fill the
        viewbox, which in turn will catch the wanted viewbox drag events.

        We ignore this event instead as it's not really used.
        """
        event.ignore()


class TimeAxisItem(DateAxisItem):
    axisDoubleClicked = Signal()

    axisStyle = {
        "autoExpandTextSpace": False,
        "tickTextWidth": 50,
        "tickTextHeight": 24,
        "tickFont": get_axis_font(size=10)}

    def __init__(self, orientation):
        super().__init__(orientation)
        self.setStyle(**self.axisStyle)

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.axisDoubleClicked.emit()
            event.accept()
            return
        super().mouseDoubleClickEvent(event)

    def mouseDragEvent(self, event):
        """Reimplemented function of PyQt"""
        event.ignore()

    def mouseClickEvent(self, event):
        """Reimplemented function of PyQt"""
        event.ignore()


class StateAxisItem(AxisItem):
    """The State Axis Item for displaying Karabo States as major ticks"""

    axisStyle = {"autoExpandTextSpace": True,
                 "tickFont": get_axis_font(size=9)}

    def tickStrings(self, values, scale, spacing):
        """Return the state names as a function of integers values"""
        return [get_state_string(value) for value in values]


class AlarmAxisItem(AxisItem):
    """The Alarm Axis Item for displaying Karabo Alarms as major ticks"""

    axisStyle = {"autoExpandTextSpace": True,
                 "tickFont": get_axis_font(size=9)}

    def tickStrings(self, values, scale, spacing):
        """Return the alarm names as a function of integers values"""
        return [get_alarm_string(value) for value in values]


class AuxPlotAxisItem(AxisItem):
    """The AxisItem for the aux plots in the image widgets"""
    axisStyle = {"autoExpandTextSpace": False,
                 "tickFont": get_axis_font(size=9)}

    def __init__(self, orientation, showValues=True):
        super().__init__(orientation, showValues)

    # ---------------------------------------------------------------------
    # PyQtGraph methods

    def _updateWidth(self):
        """Reimplemented function to match known width of main plot"""
        if self.style["showValues"]:
            self.setFixedWidth(Y_AXIS_WIDTH)
            self.picture = None
        else:
            super()._updateWidth()

    def _updateHeight(self):
        """Reimplemented function to match known height of main plot"""
        if self.style["showValues"]:
            height = X_AXIS_HEIGHT
            # Reduce space between label and axis for x-axis
            if self.orientation == "bottom":
                height -= 10
            self.setFixedHeight(height)
            self.picture = None
        else:
            super()._updateHeight()


def create_axis_items(axis=AxisType.Classic, axes_with_ticks=[]):
    """Create the basic axis items for graph plots

    The ``axis`` type defines the possible axis.
    """

    def _intersect(axis, tick_axis):
        return set(axis).intersection(tick_axis)

    time_axis, state_axis, aux_axis, alarm_axis = [], [], [], []

    if axis is AxisType.Time:
        time_axis = _intersect(AXIS_X, axes_with_ticks)
    elif axis is AxisType.State:
        state_axis = _intersect(AXIS_Y, axes_with_ticks)
        time_axis = _intersect(AXIS_X, axes_with_ticks)
    elif axis is AxisType.Alarm:
        alarm_axis = _intersect(AXIS_Y, axes_with_ticks)
        time_axis = _intersect(AXIS_X, axes_with_ticks)
    elif axis is AxisType.AuxPlot:
        aux_axis = axes_with_ticks
    items = {}
    for orientation in AXIS_ITEMS:
        has_ticks = orientation in axes_with_ticks

        if orientation in time_axis:
            axis_item = TimeAxisItem(orientation)
        elif orientation in state_axis:
            axis_item = StateAxisItem(orientation)
            axis_item.setTickSpacing(major=1, minor=1)
        elif orientation in alarm_axis:
            axis_item = AlarmAxisItem(orientation)
            axis_item.setTickSpacing(major=1, minor=1)
        elif orientation in aux_axis:
            axis_item = AuxPlotAxisItem(orientation, has_ticks)
        else:
            axis_item = AxisItem(orientation, has_ticks)

        items[orientation] = axis_item

    return items



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial

import natsort
from qtpy.QtWidgets import QAction, QDialog, QMenu, QMessageBox
from traits.api import Bool, Instance, Property, on_trait_change

import karabogui.icons as icons
from karabo.common.api import walk_traits_object
from karabo.common.project.api import DeviceServerModel
from karabogui.access import (
    AccessRole, access_role_allowed, get_access_level_for_role)
from karabogui.binding.api import ProxyStatus
from karabogui.dialogs.api import LogDialog
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.indicators import get_project_server_status_icon
from karabogui.itemtypes import ProjectItemTypes
from karabogui.logger import get_logger
from karabogui.project.dialog.server_handle import ServerHandleDialog
from karabogui.project.topo_listener import SystemTopologyListener
from karabogui.project.utils import add_device_to_server
from karabogui.singletons.api import get_manager, get_topology
from karabogui.util import move_to_cursor, version_compatible

from .bases import BaseProjectGroupController, ProjectControllerUiData

ACCESS_LEVEL_TOOLTIP = "Requires minimum '{}' access level"
SERVER_OFFLINE_TOOLTIP = "Server is offline"


class DeviceServerController(BaseProjectGroupController):
    """ A controller for DeviceServerModel objects
    """
    # Redefine model with the correct type
    model = Instance(DeviceServerModel)
    # Is the server currently online?
    online = Property(Bool, depends_on=['model'])
    # An object which listens to system topology updates
    topo_listener = Instance(SystemTopologyListener)

    def context_menu(self, project_controller, parent=None):
        # Enable/Disable based on online status
        online = self.online

        project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
        service_allowed = access_role_allowed(AccessRole.SERVICE_EDIT)
        project_edit_access_level = get_access_level_for_role(
            AccessRole.PROJECT_EDIT)
        project_allowed_tooltip = ACCESS_LEVEL_TOOLTIP.format(
            project_edit_access_level)

        service_edit_access_level = get_access_level_for_role(
            AccessRole.SERVICE_EDIT)
        service_allowed_tooltip = ACCESS_LEVEL_TOOLTIP.format(
            service_edit_access_level)
        menu = QMenu(parent)
        menu.setToolTipsVisible(True)
        edit_action = QAction(icons.edit, 'Edit', menu)
        edit_action.triggered.connect(partial(self._edit_server,
                                              parent=parent))
        edit_action.setEnabled(project_allowed)
        if not project_allowed:
            edit_action.setToolTip(project_allowed_tooltip)
        delete_action = QAction(icons.delete, 'Delete', menu)
        delete_action.triggered.connect(partial(self._delete_server,
                                                project_controller,
                                                parent=parent))
        delete_action.setEnabled(project_allowed)
        if not project_allowed:
            delete_action.setToolTip(project_allowed_tooltip)
        shutdown_action = QAction(icons.kill, 'Shutdown', menu)
        shutdown_action.triggered.connect(partial(self._shutdown_server,
                                                  parent=parent))
        shutdown_action.setEnabled(online and service_allowed)
        if not online:
            shutdown_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not service_allowed:
            shutdown_action.setToolTip(service_allowed_tooltip)
        add_action = QAction(icons.add, 'Add device', menu)
        add_action.setEnabled(online and project_allowed)
        if not online:
            add_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not project_allowed:
            add_action.setToolTip(project_allowed_tooltip)
        add_action.triggered.connect(partial(self._add_device, parent=parent))

        instantiate_all_action = QAction(
            icons.run, 'Instantiate all devices', menu)
        instantiate_all_action.setEnabled(online and service_allowed)
        if not online:
            instantiate_all_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not service_allowed:
            instantiate_all_action.setToolTip(service_allowed_tooltip)
        instantiate_all_action.triggered.connect(
            partial(self.instantiate_devices,
                    parent=parent))

        shutdown_all_action = QAction(
            icons.kill, 'Shutdown all devices', menu)
        shutdown_all_action.triggered.connect(partial(self._shutdown_devices,
                                                      parent=parent))
        shutdown_all_action.setEnabled(online and service_allowed)
        if not online:
            shutdown_all_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not service_allowed:
            shutdown_all_action.setToolTip(service_allowed_tooltip)
        remove_all_action = QAction(icons.delete, 'Delete all devices', menu)
        remove_all_action.triggered.connect(partial(self._delete_all_devices,
                                                    parent=parent))
        remove_all_action.setEnabled(project_allowed)
        if not project_allowed:
            instantiate_all_action.setToolTip(project_allowed_tooltip)

        show_action = QAction(icons.yes, 'Select in topology', menu)
        show_action.triggered.connect(self._show_server)
        show_action.setVisible(self.online)

        menu.addAction(edit_action)
        menu.addAction(delete_action)
        menu.addAction(shutdown_action)
        menu.addSeparator()
        menu.addAction(add_action)
        menu.addAction(instantiate_all_action)
        menu.addAction(shutdown_all_action)
        menu.addAction(remove_all_action)

        # Sub menu for device sorting
        sort_menu = menu.addMenu("&Sort options")
        sort_menu.setIcon(icons.edit)
        sort_menu.setEnabled(project_allowed)
        if not project_allowed:
            sort_menu.setToolTip(project_allowed_tooltip)

        sort_alpha = QAction(
            icons.stringAttribute, 'Sort devices alphabetically', menu)
        sort_alpha.triggered.connect(partial(self._sort_alphabetically,
                                             parent=parent))

        sort_domain = QAction(icons.folderDomain,
                              'Sort devices by domain', menu)
        sort_domain.triggered.connect(partial(self._sort_devices_naming,
                                              level=0, parent=parent))
        sort_type = QAction(icons.folderType,
                            'Sort devices by type', menu)
        sort_type.triggered.connect(partial(self._sort_devices_naming,
                                            level=1, parent=parent))
        sort_member = QAction(icons.deviceInstance,
                              'Sort devices by member', menu)
        sort_member.triggered.connect(partial(self._sort_devices_naming,
                                              level=2, parent=parent))

        sort_menu.addAction(sort_alpha)
        sort_menu.addAction(sort_domain)
        sort_menu.addAction(sort_type)
        sort_menu.addAction(sort_member)

        menu.addAction(show_action)

        return menu

    def create_ui_data(self):
        ui_data = ProjectControllerUiData()
        self._update_icon(ui_data)
        return ui_data

    def system_topology_callback(self, devices, servers):
        """ This callback is called by the ``SystemTopologyListener`` object
        in the ``topo_listener`` trait.
        """
        for server_id, _, status in servers:
            # NOTE: the host is not checked here since it might ot have been
            # set yet and furthermore the server_id is unique in the system
            if self.model.server_id == server_id:
                self.model.status = status

    def _get_display_name(self):
        """Traits property getter for ``display_name``
        """
        return self.model.server_id

    def info(self):
        return {'type': ProjectItemTypes.SERVER,
                'serverId': self.model.server_id,
                'host': self.model.host}

    def delete_press(self, project_controller, parent=None):
        """Reimplemented function on `BaseProjectController`"""
        self._delete_server(project_controller, parent)

    def double_click(self, project_controller, parent=None):
        """Reimplemented function on `BaseProjectController`"""
        server_id = self.model.server_id
        if not self.online:
            get_logger().info(
                f"Cannot retrieve logs for device server <b>{server_id}</b>, "
                "it is <b>offline</b>.")
            return

        attrs = get_topology().get_attributes(f"server.{server_id}")
        assert attrs is not None

        version = attrs.get("karaboVersion", "0.0.0")
        if not version_compatible(version, 2, 14):
            get_logger().info(
                f"Cannot retrieve logs for device server <b>{server_id}</b>. "
                f"Retrieval of logs is available from version 2.14.0.")
            return

        widget = LogDialog(server_id, parent=parent)
        move_to_cursor(widget)
        widget.show()
        widget.raise_()
        widget.activateWindow()

    # ----------------------------------------------------------------------
    # traits handlers

    @on_trait_change("model.modified,model:server_id")
    def update_ui_label(self):
        """ Whenever the project is modified it should be visible to the user
        """
        self._update_icon(self.ui_data)

    @on_trait_change("model.status")
    def status_change(self):
        icon = get_project_server_status_icon(self.model.status)
        if icon is not None:
            self.ui_data.icon = icon

    def _get_online(self):
        return self.model.status is not ProxyStatus.OFFLINE

    def _topo_listener_changed(self, name, old, new):
        """Handle broadcast event registration/unregistration here.
        """
        if old is not None:
            unregister_from_broadcasts(
                {KaraboEvent.SystemTopologyUpdate: old._event_topology})
        if new is not None:
            register_for_broadcasts(
                {KaraboEvent.SystemTopologyUpdate: new._event_topology})

    # ----------------------------------------------------------------------
    # Util methods

    def _update_icon(self, ui_data):
        # Get current status of server
        self.model.status = _get_server_status(self.model.server_id)
        icon = get_project_server_status_icon(self.model.status)
        ui_data.icon = icon

    # ----------------------------------------------------------------------
    # action handlers

    def _show_server(self):
        serverId = self.model.server_id
        broadcast_event(KaraboEvent.ShowDevice,
                        {'deviceId': serverId, 'showTopology': True})

    def _add_device(self, parent=None):
        """Add a device instance to the server
        """
        # NOTE: This can't be connected to the QAction signal, because then
        # a boolean will be passed as the class_id argument
        add_device_to_server(self.model, parent=parent)

    def _delete_server(self, project_controller, parent=None):
        """ Remove the macro associated with this item from its project
        """
        server = self.model
        ask = ('Are you sure you want to delete \"<b>{}</b>\".<br /> '
               'Continue action?'.format(server.simple_name))
        msg_box = QMessageBox(QMessageBox.Question, 'Delete server',
                              ask, QMessageBox.Yes | QMessageBox.No,
                              parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.Yes:
            project = project_controller.model
            if server in project.servers:
                del server.devices[:]
                project.servers.remove(server)

    def _edit_server(self, parent=None):
        dialog = ServerHandleDialog(self.model, parent=parent)
        move_to_cursor(dialog)
        result = dialog.exec()
        if result == QDialog.Accepted:
            self.model.server_id = dialog.server_id

    def _shutdown_server(self, parent=None):
        server = self.model
        get_manager().shutdownServer(server.server_id, parent=parent)

    def _delete_all_devices(self, parent=None):
        server = self.model
        ask = ('Do you really want to delete all devices of <br> server '
               '<b>{}</b>?').format(server.simple_name)
        msg_box = QMessageBox(QMessageBox.Question, 'Delete all devices', ask,
                              QMessageBox.Yes | QMessageBox.No, parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.No:
            return

        server.devices[:] = []

    def instantiate_devices(self, parent=None):
        """Public action handler to instantiate all devices from this server

        This action handler is also publicly used by the project controller
        """
        server = self.model
        for dev_inst_item in self.children:
            dev_inst_item.instantiate(server, parent)

    def _shutdown_devices(self, parent=None):
        server = self.model
        ask = ('Do you really want to shutdown all devices of <br> server '
               '<b>{}</b>?').format(server.simple_name)
        msg_box = QMessageBox(QMessageBox.Question, 'Shutdown all devices',
                              ask, QMessageBox.Yes | QMessageBox.No,
                              parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.No:
            return

        for dev_inst_item in self.children:
            dev_inst_item.shutdown_device(show_confirm=False)

    def _sort_alphabetically(self, parent=None):
        server_model = self.model
        devices = list(server_model.devices)
        devices = natsort.natsorted(devices,
                                    key=lambda device: device.instance_id)
        del server_model.devices[:]
        server_model.devices.extend(devices)

    def _sort_devices_naming(self, level=0, parent=None):
        server_model = self.model
        devices = list(server_model.devices)

        def sort_func(device):
            device_id = device.instance_id
            karabo_name = device_id.split("/")
            return device_id if len(karabo_name) != 3 else karabo_name[level]

        devices = natsort.natsorted(devices, key=sort_func)
        del server_model.devices[:]
        server_model.devices.extend(devices)


# ----------------------------------------------------------------------

def _get_server_status(server_id):
    topology = get_topology()
    attributes = topology.get_attributes(f'server.{server_id}')
    if attributes is not None:
        return ProxyStatus.ONLINE
    return ProxyStatus.OFFLINE


def get_project_servers(project_controller):
    """Given a ``project_controller`` return all the online and offline servers
    """
    online = []
    offline = []

    def visitor(obj):
        if isinstance(obj, DeviceServerController):
            if obj.online:
                online.append(obj)
            else:
                offline.append(obj)

    walk_traits_object(project_controller, visitor)

    return online, offline



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import LabelItem

from karabogui.graph.common.api import KaraboLegend, float_to_string


class ScaleLegend(KaraboLegend):

    def __init__(self):
        super().__init__()
        self._label = LabelItem(color='w', size="8pt")
        self.layout.addItem(self._label, 0, 0)
        self.layout.setContentsMargins(2, 2, 2, 2)

    def set_value(self, x_scale, y_scale, x_units, y_units):
        x_scale, y_scale = (float_to_string(num) for num in [x_scale, y_scale])
        self._label.setText("<b>Scale</b><br>"
                            "x: {} {}<br>"
                            "y: {} {}".format(x_scale, x_units,
                                              y_scale, y_units))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QMessageBox

from karabo.common.scenemodel.api import DisplayCommandModel
from karabogui.controllers.display.command import DisplayCommand
from karabogui.dialogs.api import ProxiesDialog
from karabogui.sceneview.widget.api import ControllerContainer
from karabogui.testing import click_button


def test_proxies_dialog(gui_app, mocker):
    model = DisplayCommandModel(keys=["TestDevice.hello", "TestDevice.bye"])
    container = ControllerContainer(DisplayCommand, model, None)

    controller = container.widget_controller
    assert controller.widget is not None
    assert len(controller._actions) == 2

    assert container.toolTip() == "TestDevice.hello, TestDevice.bye"
    dialog = ProxiesDialog(container)
    mbox = mocker.patch("karabogui.dialogs.proxies_dialog.QMessageBox")
    mbox.return_value = QMessageBox.Yes
    view = dialog.ui_list_view
    model = view.model()
    view.setCurrentIndex(model.index(0, 0))
    click_button(dialog.ui_remove_button)
    # A button is removed
    assert len(controller._actions) == 1
    assert container.toolTip() == "TestDevice.hello"
    container.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re
from pathlib import Path

import yaml
from qtpy.QtCore import QProcess
from qtpy.QtWidgets import QApplication

from karabogui.singletons.api import get_config, get_network
from karabogui.util import process_qt_events

LINUX_DESKTOP_FILE_TEMPLATE = """\
[Desktop Entry]
Name={name}
Exec={command}
Icon={icon}
Terminal=false
Type=Application
MimeType=x-scheme-handler/{scheme};
"""


def create_linux_desktop_file(name, command, icon, scheme):
    # Resolve path
    path = Path.home() / ".local" / "share" / "applications"
    path.mkdir(parents=True, exist_ok=True)

    path /= get_valid_filename(name).lower() + ".desktop"
    # Write desktop file
    with path.open("w") as f:
        contents = LINUX_DESKTOP_FILE_TEMPLATE.format(
            name=name, command=command, icon=icon, scheme=scheme)
        f.write(contents)

    return path


def get_valid_filename(s):
    """
    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    >>> get_valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'

    From Django:
    https://github.com/django/django/blob/master/django/utils/text.py
    """
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)


def close_app():
    app = QApplication.instance()
    if app is not None:
        process_qt_events(app, timeout=1000)
        app.quit()


def save_concert_file(file_name, scene_data):
    """ Write the scene information to a yaml file, for karabo-concert.
    """
    domain = get_config()["domain"]
    network = get_network()
    hostname = network.hostname
    port = network.port

    scenes = []
    for uuid, properties in scene_data.items():
        x = properties["x"]
        y = properties["y"]
        scenes.append({"uuid": uuid, "x": x, "y": y})

    concert = {"domain": domain, "host": hostname,
               "port": port, "scenes": scenes}
    with open(file_name, "w") as yaml_file:
        yaml.dump(concert, yaml_file)


def run_concert(file_name):
    """Start karabo-concert as a separate process"""
    QProcess.startDetached(f"karabo-concert {file_name}")



from qtpy.QtCore import QRect, Qt, Slot
from qtpy.QtWidgets import QAction, QDialog, QMenu, QToolButton

from karabogui import icons
from karabogui.dialogs.api import PopupButtonDialog
from karabogui.util import move_to_cursor
from karabogui.widgets.api import TextPopupWidget
from karabogui.widgets.hints import KaraboSceneWidget


def _get_icon_data():
    # Do not make this as global variable, as it gets loaded before the
    # icons are initalized in the tests.
    return {
        "General Info": icons.note_info,
        "Help Info": icons.note_warning,
        "Critical Info": icons.note_alert}


def _get_icon(info_type):
    data = _get_icon_data()
    icon = data.get(info_type, icons.note_info)
    return icon


class PopupButtonWidget(KaraboSceneWidget, QToolButton):
    """A button that shows a title and detailed text as popup on
    mouse click"""
    def __init__(self, model, parent=None):
        super().__init__(model=model, parent=parent)
        self.setMinimumSize(self.iconSize())

        edit_text_action = QAction("Edit Text", self)
        edit_text_action.triggered.connect(self._edit_text)
        self.addAction(edit_text_action)

        menu = QMenu(parent=self)
        info_type_action = QAction(menu)
        info_type_action.setText("Info Type")

        for text, icon in _get_icon_data().items():
            action = QAction(icon, text, menu)
            action.triggered.connect(self._update_button_icon)
            menu.addAction(action)
        info_type_action.setMenu(menu)
        self.addAction(info_type_action)

        self.set_widget_properties(model)

        self.setGeometry(QRect(model.x, model.y,
                               model.width, model.height))
        self.setStyleSheet("QToolButton { border: none; }"
                           "QToolButton::menu-indicator {image:none;}")

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    def set_widget_properties(self, model):
        icon = _get_icon(model.info_type)
        self.setIcon(icon)
        self.setIconSize(self.size())

        tooltip = f"{model.info_type}: Click for details"
        self.setToolTip(tooltip)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            widget = TextPopupWidget(self.model, parent=self)
            move_to_cursor(widget)
            widget.show()
            event.accept()

    def resizeEvent(self, event):
        """Override to set the icon size to match the widget size."""
        super().resizeEvent(event)
        self.setIconSize(self.size())

    def edit(self, scene_view):
        """Override to open a dialog for editing the sticker."""
        dialog = PopupButtonDialog(self.model, parent=scene_view)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Rejected:
            return

        model = dialog.model
        self.model.trait_set(text=model.text,
                             popup_width=model.popup_width,
                             popup_height=model.popup_height)
        self.set_widget_properties(model)

    @Slot()
    def _edit_text(self):
        self.edit(scene_view=self.parent())

    @Slot()
    def _update_button_icon(self):
        action = self.sender()
        if action:
            self.model.trait_set(info_type=action.text())
            self.set_widget_properties(self.model)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import Point
from qtpy.QtCore import QRectF, Slot
from qtpy.QtGui import QPainterPath, QPainterPathStroker

from karabogui.graph.common.utils import float_to_string

from .base import KaraboROI
from .utils import ImageRegion, set_roi_html


class CrosshairROI(KaraboROI):

    def __init__(self, pos=None, size=(1, 1), name='',
                 scaleSnap=False, translateSnap=False, pen=None):
        """Reimplementation of pyqtgraph CrosshairROI to support transforms
        and add a few fixes/performance improvements:

        - Force redraw the crosshair shape with sigRegionChanged"""

        super().__init__(pos, size, name, scaleSnap, translateSnap, pen=pen)
        self._shape = None
        self._path = None
        self.sigRegionChanged.connect(self.redraw)
        self.aspectLocked = True

    @Slot()
    def _change_roi_text_item_details(self):
        if not self.textItem:
            return

        center = self.center
        self.textItem.setPos(center)
        x, y = (float_to_string(coord) for coord in center)
        self.textItem.setHtml(set_roi_html(name=self.name, center=(x, y)))

    def invalidate(self):
        self._shape = None
        self.prepareGeometryChange()

    @property
    def center(self):
        return self.pos()

    @property
    def coords(self):
        return tuple(self._absolute_position)

    # ---------------------------------------------------------------------
    # Public methods

    def get_region(self, imageItem):
        image = imageItem.image
        x, y = np.floor(self._absolute_position).astype(int)

        if 0 <= x < image.shape[1]:
            y_region = image[:, x]
            y_slice = slice(image.shape[0])
        else:
            y_region = np.array([])
            y_slice = slice(0)

        if 0 <= y < image.shape[0]:
            x_region = image[y, :]
            x_slice = slice(image.shape[1])
        else:
            x_region = np.array([])
            x_slice = slice(0)

        image_region = ImageRegion([x_region, y_region], ImageRegion.Line,
                                   x_slice, y_slice)

        return image_region

    def redraw(self):
        """Conventional Qt items use self.update() to redraw/repaint.
           CrosshairROI, on the other hand, needs to recalculate the paths
           prior to repainting. This can be done using self.invalidate()"""
        self.invalidate()

    # ---------------------------------------------------------------------
    # PyQt methods

    def boundingRect(self):
        shape = self.shape()
        return shape.boundingRect() if shape is not None else QRectF()

    def paint(self, p, *args):
        p.setPen(self.currentPen)
        p.drawPath(self._path)

    def shape(self):
        if self._shape is None:
            self._path = self._create_path()
            self._shape = self._create_shape(self._path)

        return self._shape

    def viewTransformChanged(self):
        super().viewTransformChanged()
        self.invalidate()

    # ---------------------------------------------------------------------
    # PyQtGraph methods

    def getSnapPosition(self, pos, snap=None):
        """Patching to support snapping based on scaling"""
        scaled_pos = np.floor(np.array(pos) / self._scaling) * self._scaling
        return Point(*scaled_pos)

    # ---------------------------------------------------------------------
    # Private methods

    def _create_shape(self, path):
        """Creates the stroke from the input path. When the viewbox is not
           rendered yet, which happens during initialization,
           `mapToDevice` returns None"""
        path = self.mapToDevice(path)
        if path is not None:
            stroker = QPainterPathStroker()
            stroker.setWidth(10)
            outline = stroker.createStroke(path)
            return self.mapFromDevice(outline)
        return None

    def _create_path(self):
        # Factor is the desired width/height wrt viewbox divided by 2
        factor = 0.1 if self.selected else 0.05
        width, height = self._get_effective_dimensions()

        # Create the path from the half dimension
        path = QPainterPath()
        # paint x-axis
        half_width = width * factor
        path.moveTo(Point(-half_width, 0))
        path.lineTo(Point(half_width, 0))
        # paint y-axis
        half_height = height * factor
        path.moveTo(Point(0, -half_height))
        path.lineTo(Point(0, half_height))

        return path

    def _get_effective_dimensions(self):
        viewbox = self.getViewBox()
        rect = self.mapRectFromView(viewbox.viewRect())
        width, height = rect.width(), rect.height()

        # Get factor that generates least common multiple from the width/height
        min_dim, max_dim = width, height
        if min_dim > max_dim:
            min_dim, max_dim = max_dim, min_dim
        factor = max_dim // min_dim

        # Get effective dimension from the multiple factor
        if width < height:
            height = width * factor
        else:
            width = height * factor

        # Correct dimension with the screen geometry ratio (rect of the plot)
        geom = viewbox.screenGeometry()
        ratio = geom.height() / geom.width()
        if ratio > 1:
            height /= ratio
        else:
            width *= ratio

        return width, height



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QMenu
from traits.api import (
    ArrayOrNone, Bool, Dict, Instance, Property, Type, on_trait_change)

from ..base.controller import BaseController, ControllerAggregator
from .analyzer import ProfileAnalyzer
from .plot import ProfilePlot
from .stats import ProfileStats


class ProfileController(BaseController):

    # The analyzer class for the profile plot
    analyzer = Instance(ProfileAnalyzer, args=())

    # The plot item controller class for the profile plot
    plot_klass = Type(ProfilePlot)

    # Fit the profile data. This is toggled by a context menu.
    fitted = Bool(False)

    # Transformed image axis data. From the analyzer class
    axis = Property(ArrayOrNone)

    # Smoothen the resulting signal. From the analyzer class.
    smooth = Property

    # -----------------------------------------------------------------------
    # Public methods

    def process(self, region, axis=0):
        """Process is made up of two parts:
           1. Analyze the data first
           2. Plot the data"""

        # Clean up first
        self.clear_data()

        profiles = self.analyzer.analyze(region, axis=axis)
        if profiles is None:
            return
        self.plot.set_data(*profiles)
        if self.fitted:
            x_fit, y_fit = self.analyzer.fit()
            self.plot.set_fit(x_fit, y_fit)

    # -----------------------------------------------------------------------
    # Trait handlers

    def _fitted_changed(self, fitted):
        x_fit, y_fit = [], []
        if fitted:
            x_fit, y_fit = self.analyzer.fit()
        self.plot.set_fit(x_fit, y_fit)

    # -----------------------------------------------------------------------
    # Trait properties

    def _set_axis(self, axis):
        self.analyzer.axis_data = axis

    def _get_smooth(self):
        return self.analyzer.smooth

    def _set_smooth(self, is_smooth):
        self.analyzer.smooth = is_smooth


class ProfileAggregator(ControllerAggregator):

    # The profile aux plot contains two profile plots
    contents = Dict({"top": ProfileController, "left": ProfileController})

    # Profile plot instances that are connected for x- and y-axis
    x_plot = Property(Instance(ProfileController))
    y_plot = Property(Instance(ProfileController))

    # --- Convenience properties to get/set values on the histogram plot ---
    # Smoothen the resulting signal. Retrieved from both profile plots
    smooth = Property

    def __init__(self, **traits):
        super().__init__(**traits)

        # Create shared menu
        menu = QMenu()
        enable_fit_action = QAction("Gaussian fitting", menu)
        enable_fit_action.setCheckable(True)
        enable_fit_action.setChecked(self.show_stats)
        enable_fit_action.triggered.connect(self._enable_fitting)
        menu.addAction(enable_fit_action)

        self.x_plot.plot.menu = menu
        self.y_plot.plot.menu = menu

    # -----------------------------------------------------------------------
    # Public methods

    def process(self, region):
        self.x_plot.process(region, axis=0)
        self.y_plot.process(region, axis=1)

        if self.show_stats:
            self.stats = ProfileStats(x_stats=self.x_plot.analyzer.stats,
                                      y_stats=self.y_plot.analyzer.stats)

    def link_view(self, view):
        """Connects the image plot viewbox to the respective aux plots"""
        self.x_plot.link_view(axis=0, view=view)
        self.y_plot.link_view(axis=1, view=view)

    def set_axes(self, x_data, y_data):
        self.x_plot.axis = x_data
        self.y_plot.axis = y_data

    def destroy(self):
        super().destroy()
        self.on_trait_change(self._show_stats, "show_stats", remove=True)

    # -----------------------------------------------------------------------
    # Private methods

    def _enable_fitting(self, enabled):
        self.x_plot.fitted = self.y_plot.fitted = enabled
        self.show_stats = enabled

    # -----------------------------------------------------------------------
    # Trait handlers

    @on_trait_change("show_stats")
    def _show_stats(self, show_stats):
        stats = None
        if show_stats:
            stats = ProfileStats(x_stats=self.x_plot.analyzer.stats,
                                 y_stats=self.y_plot.analyzer.stats)
        self.stats = stats

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_x_plot(self):
        return self.controllers["top"]

    def _get_y_plot(self):
        return self.controllers["left"]

    def _get_smooth(self):
        return self.x_plot.smooth and self.y_plot.smooth

    def _set_smooth(self, smooth):
        self.x_plot.smooth = smooth
        self.y_plot.smooth = smooth



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtWidgets import QAction, QInputDialog
from traits.api import ArrayOrNone, Instance, WeakRef

from karabo.common.scenemodel.api import (
    VectorScatterGraphModel, build_graph_config, build_model_config)
from karabogui import icons
from karabogui.binding.api import (
    PropertyProxy, VectorNumberBinding, get_binding_value)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.common.toolbar.widgets import create_button
from karabogui.graph.plots.api import KaraboPlotView, ScatterGraphPlot
from karabogui.graph.plots.dialogs.data_analysis import DataAnalysisDialog

MIN_POINT_SIZE = 0.1
MAX_POINT_SIZE = 10.0


@register_binding_controller(ui_name='Vector XY Scatter',
                             klassname='VectorScatterGraph',
                             binding_type=VectorNumberBinding,
                             can_show_nothing=False)
class DisplayVectorScatterGraph(BaseBindingController):
    """The scatter graph widget provides a scatter plot from property proxies
    """
    model = Instance(VectorScatterGraphModel, args=())

    # Internal traits
    _x_proxy = Instance(PropertyProxy)
    _y_proxy = Instance(PropertyProxy)
    _last_x_value = ArrayOrNone
    _plot = WeakRef(ScatterGraphPlot)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.add_cross_target()
        toolbar = widget.add_toolbar()
        widget.enable_export()
        widget.stateChanged.connect(self._change_model)

        self._plot = widget.add_scatter_item(cycle=False)
        # assigning proxy is safe and wanted here!
        self._x_proxy = self.proxy
        self._last_x_value = get_binding_value(self.proxy)
        widget.restore(build_model_config(self.model))
        self._plot.setSize(self.model.psize)

        point_size_action = QAction("Point Size", widget)
        point_size_action.triggered.connect(self._configure_point_size)
        point_size_action.setIcon(icons.scatter)
        widget.addAction(point_size_action)
        button = create_button(
            icon=icons.data_analysis, checkable=False, tooltip="Data Analysis",
            on_clicked=self.show_data_analysis_dialog)
        toolbar.add_button(button)

        return widget

    # ----------------------------------------------------------------

    def show_data_analysis_dialog(self):
        config = build_graph_config(self.model)
        dialog = DataAnalysisDialog(
            proxies=self.proxies[1:], config=config, baseline_proxy=self.proxy,
            parent=self.widget)
        dialog.show()

    def add_proxy(self, proxy):
        if self._y_proxy is None:
            self._y_proxy = proxy
            return True

        return False

    def remove_proxy(self, proxy):
        if self._y_proxy is not None:
            self._y_proxy = None
            return True

        return False

    def value_update(self, proxy):
        value = proxy.value
        if proxy is self._x_proxy:
            self._last_x_value = value
        elif proxy is self._y_proxy:
            if value is None:
                return
            if self._last_x_value is not None and self.widget is not None:
                size_x = len(self._last_x_value)
                size_y = len(value)
                size = min(size_x, size_y)
                if size == 0:
                    self._plot.setData([], [])
                    return

                x = self._last_x_value
                if size_x != size_y:
                    x = self._last_x_value[:size]
                    value = value[:size]
                self._plot.setData(x, value)

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)

    def _configure_point_size(self):
        psize, ok = QInputDialog.getDouble(self.widget, 'Size of points',
                                           'Pointsize:', self.model.psize,
                                           MIN_POINT_SIZE, MAX_POINT_SIZE)
        if ok:
            self._plot.setSize(psize)
            self.model.psize = psize



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on October 9, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

import weakref
from collections import deque

from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget

from karabogui import icons

from .utils import get_panel_ui


class SearchBar(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_panel_ui("project_filter.ui"), self)

        self.label_filter.textChanged.connect(self._filter_changed)
        self.label_filter.returnPressed.connect(self._arrow_right_clicked)
        self.arrow_left.setIcon(icons.arrowLeft)
        self.arrow_left.clicked.connect(self._arrow_left_clicked)
        self.arrow_right.setIcon(icons.arrowRight)
        self.arrow_right.clicked.connect(self._arrow_right_clicked)
        self.case_sensitive.clicked.connect(self._update_filter)
        self.regexp.clicked.connect(self._update_filter)

        # The list of found nodes
        self.found = []
        # A deque array indicates the current selection in `self.found`
        self.index_array = deque([])
        self.reset()

        self.model = None

    def setModel(self, model):
        handler_method = getattr(model, 'findNodes', None)
        if handler_method is None or not callable(handler_method):
            msg = ('A model must implement a "findNodes" method '
                   'if it is attached for the search bar!')
            raise RuntimeError(msg)

        handler_method = getattr(model, 'selectNode', None)
        if handler_method is None or not callable(handler_method):
            msg = ('A model must implement a "selectNode" method '
                   'if it is attached for the search bar!')
            raise RuntimeError(msg)

        self.model = weakref.ref(model)

    # -----------------------------------------
    # Public interface

    def reset(self, enable=False):
        """Reset the model search bar"""
        self.found = []
        self.index_array = deque([])
        self.label_filter.setText("")
        self.label_filter.setEnabled(enable)
        self.case_sensitive.setEnabled(enable)
        self.regexp.setEnabled(enable)
        self.arrow_left.setEnabled(False)
        self.arrow_right.setEnabled(False)

    # -----------------------------------------

    def _select_node(self):
        """Pick the first number in index array, select the corresponding node
        """
        if self.model is None:
            raise RuntimeError("Model is not set!")

        idx = next(iter(self.index_array))
        node = self.found[idx]
        parent = node.parent
        # NOTE: The node might have left already!
        if parent is not None and node in parent.children:
            self.model().selectNode(node)

    # -----------------------------------------
    # Qt Slots

    @Slot(str)
    def _filter_changed(self, text):
        """ Slot is called whenever the search filter text was changed"""
        if self.model is None:
            raise RuntimeError("Model is not set!")

        if text:
            model = self.model()
            kwargs = {'case_sensitive': self.case_sensitive.isChecked(),
                      'use_reg_ex': self.regexp.isChecked()}
            self.found = model.findNodes(text, **kwargs)
        else:
            self.found = []

        n_found = len(self.found)
        self.index_array = deque(range(n_found))
        if self.index_array:
            result = f"{n_found} Results"
            self._select_node()
        else:
            result = "No results"
        self.result.setText(result)

        enable = n_found > 0
        self.arrow_left.setEnabled(enable)
        self.arrow_right.setEnabled(enable)

    @Slot()
    def _update_filter(self):
        text = self.label_filter.text()
        self._filter_changed(text)

    @Slot()
    def _arrow_left_clicked(self):
        """rotate index array to the right, so the first index point to the
        previous found node
        """
        self.index_array.rotate(1)
        self._select_node()

    @Slot()
    def _arrow_right_clicked(self):
        """rotate index array to the left, so the first index point to the
        next found node
        """
        # Enter key press is linked to arrow right, we have to validate if
        # there are results to show
        if self.index_array:
            self.index_array.rotate(-1)
            self._select_node()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on January 16, 2014
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QSize
from qtpy.QtWidgets import QSizePolicy, QToolBar, QWidget


class ToolBar(QToolBar):
    def __init__(self, title="", parent=None):
        super().__init__(title, parent=parent)

        iconSize = QSize(32, 32)
        iconSize *= 0.6
        self.setIconSize(iconSize)
        self.setMovable(False)

    def add_expander(self):
        """Add an empty widget to the toolbar which absorbs horizontal space.
        """
        expander = QWidget()
        expander.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.addWidget(expander)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 6, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op
import re
from functools import partial

from qtpy.QtCore import QEvent, QSize, Qt, Slot
from qtpy.QtGui import QIcon, QKeySequence, QPalette
from qtpy.QtWidgets import (
    QAction, QActionGroup, QApplication, QDialog, QMenu, QMessageBox,
    QScrollArea)

import karabogui.access as krb_access
from karabo.common.scenemodel.api import write_scene
from karabogui import icons
from karabogui.dialogs.api import ResizeSceneDialog, UserSessionDialog
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.indicators import get_topic_color
from karabogui.sceneview.api import SceneView
from karabogui.sceneview.const import QT_CURSORS, SCENE_BORDER_WIDTH
from karabogui.sceneview.tools.api import (
    ArrowSceneTool, BoxHSceneAction, BoxVSceneAction, CreateToolAction,
    DeviceSceneLinkTool, GroupEntireSceneAction, GroupSceneAction,
    ImageRendererTool, InstanceStatusTool, LineSceneTool, PopupButtonTool,
    RectangleSceneTool, SceneAlignAction, SceneBringToFrontAction,
    SceneCopyAction, SceneCutAction, SceneDeleteAction, SceneLinkTool,
    SceneMoveAction, ScenePasteAction, ScenePasteReplaceAction,
    SceneSelectAllAction, SceneSelectionTool, SceneSendToBackAction,
    StickerTool, TextSceneTool, UngroupSceneAction, WebLinkTool)
from karabogui.singletons.api import get_config, get_network
from karabogui.util import getSaveFileName, move_to_cursor
from karabogui.widgets.toolbar import ToolBar

from .base import BasePanelWidget
from .panel_info import create_scene_info

# NOTE: This is the amount of padding added by ScenePanel's QFrame parent
# We need to take it into account when undocking!!
QFRAME_PADDING = 4

ACCESS_LEVEL_INFO = """<html><head/><body><p><span style=" font-size:10pt;
font-style:italic; font-weight:normal; color:#242424;">Currently logged in as
temporary user <b>'{user}'</b> with the access level <b>'{access_level}'</b>.
Do you really want to end the Temporary session?</span></p>
"""


class ScenePanel(BasePanelWidget):
    def __init__(self, model, connected_to_server):
        # cache a reference to the scene model
        self.model = model
        self.connected_to_server_at_init = connected_to_server
        super().__init__(model.simple_name, allow_closing=True)

        # Add a handler for the window title. Don't forget to remove it later!
        self.model.on_trait_change(self.set_title, 'simple_name')

    # ----------------------------
    # BasePanelWidget Methods

    def attach_to_container(self, container):
        super().attach_to_container(container)
        self.home_tool_bar.setVisible(container is None)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        self.scene_view = SceneView(model=self.model, parent=self)
        self.scroll_widget = ResizableScrollArea(self.scene_view, parent=self)
        self.scroll_widget.setWidget(self.scene_view)
        self.scroll_widget.setBackgroundRole(QPalette.Dark)
        self.scroll_widget.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.scroll_widget.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.scroll_widget.setContentsMargins(0, 0, 0, 0)
        return self.scroll_widget

    def dock(self):
        """Called before this panel is docked into the main window.
        """
        self.scroll_widget.setWidgetResizable(False)
        self.home_tool_bar.setVisible(False)

    def undock(self, maximized=False):
        """Called before this panel is undocked from the main window.
        """
        width, height = self._compute_panel_size()
        screen_rect = QApplication.desktop().screenGeometry()
        scene_view = self.scene_view
        if (width < screen_rect.width() and height < screen_rect.height()):
            # Resize panel
            self.resize(width, height)
            # Enlarge the scene widget to its actual size
            with scene_view.ignore_resize_events():
                self.scroll_widget.setWidgetResizable(True)
        self.home_tool_bar.setVisible(True)

    def toolbars(self):
        """This should create and return one or more `ToolBar` instances needed
        by this panel.
        """
        self.create_toolbar_actions(self.connected_to_server_at_init)

        always_visible_tb = ToolBar(parent=self)
        always_visible_tb.addAction(self.ac_design_mode)
        always_visible_tb.addSeparator()
        always_visible_tb.addAction(self.ac_apply_all)
        always_visible_tb.addAction(self.ac_decline_all)
        if get_config()["development"]:
            always_visible_tb.addAction(self.ac_save_scene)

        always_visible_tb.add_expander()

        tool_bar = ToolBar("Drawing", parent=self)
        tool_bar.setVisible(self.scene_view.design_mode)
        self.scene_view.setFocusProxy(tool_bar)
        tool_bar.setFocusPolicy(Qt.StrongFocus)

        tool_bar.addSeparator()
        for action in self.qactions:
            tool_bar.addAction(action)

        self.drawing_tool_bar = tool_bar
        # Add placeholder widget to toolbar
        self.drawing_tool_bar.add_expander()

        # Reset to selection tool when no current tool is selected
        self.scene_view.resetToSelectionTool.connect(
            self.ac_selection_tool.trigger)

        home_tool_bar = ToolBar(parent=self)
        ac_show_editor = QAction(icons.homeEdit, "&Home", self)
        ac_show_editor.triggered.connect(self.show_editor)
        home_tool_bar.addSeparator()
        home_tool_bar.addAction(ac_show_editor)

        self.tbUserSession = QAction(parent=self)
        home_tool_bar.addAction(self.tbUserSession)
        self.tbUserSession.triggered.connect(self.onTemporarySession)
        home_tool_bar.addAction(self.tbUserSession)

        self.home_tool_bar = home_tool_bar
        self.always_visible_tb = always_visible_tb

        return [always_visible_tb, self.drawing_tool_bar, home_tool_bar]

    @Slot()
    def onTemporarySession(self):
        if krb_access.is_temporary_session():
            ask = ACCESS_LEVEL_INFO.format(
                user=krb_access.TEMPORARY_SESSION_USER,
                access_level=krb_access.GLOBAL_ACCESS_LEVEL.name)
            dialog = QMessageBox(
                QMessageBox.Question, 'Temporary Session', ask,
                QMessageBox.Yes | QMessageBox.No, parent=self)
            move_to_cursor(dialog)
            if dialog.exec() == QMessageBox.Yes:
                get_network().endTemporarySession()
        else:
            self._begin_temporary_session()

    def _begin_temporary_session(self):
        if not krb_access.is_authenticated():
            return

        dialog = UserSessionDialog(parent=self)
        dialog.show()

    def setTemporaryButtonVisible(self, enable: bool) -> None:
        """Make the tempsession visible depending on the boolean `enable`"""
        self.tbUserSession.setVisible(enable)

    def setSessionButton(self, icon: QIcon, tooltip: str) -> None:
        """Set the icon and tooltip for the Temporary Session button"""
        self.tbUserSession.setToolTip(tooltip)
        self.tbUserSession.setIcon(icon)

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            # Tell the scene view and scroll widget to destruct
            if self.scene_view is not None:
                self.scene_view.setParent(None)
                self.scene_view.destroy()
                self.scene_view = None
            if self.scroll_widget is not None:
                self.scroll_widget.setParent(None)
                self.scroll_widget.destroy()
                self.scroll_widget = None
            # Remove the window title handler
            self.model.on_trait_change(self.set_title, 'simple_name',
                                       remove=True)
            # Tell the world we're closing
            broadcast_event(KaraboEvent.MiddlePanelClosed,
                            {'model': self.model})

    def hideEvent(self, event):
        if self.scene_view is not None:
            self.scene_view.set_tab_visible(False)

    def showEvent(self, event):
        if self.scene_view is not None:
            self.scene_view.set_tab_visible(True)

        # When undocked, make sure our panel has an appropriate size
        if not self.is_docked:
            width, height = self._compute_panel_size()
            self.resize(width, height)
            if self.scroll_widget is not None:
                self.scroll_widget.setWidgetResizable(True)

    def set_toolbar_style(self, karabo_topic):
        color = get_topic_color(karabo_topic)
        if color is not None:
            self.toolbar.setStyleSheet(
                f"""QToolBar {{ background-color: rgba{color} }};""")

    # ----------------------------
    # Qt slots

    @Slot()
    def show_editor(self):
        broadcast_event(KaraboEvent.RaiseEditor, {"proxy": None})

    @Slot()
    def save_scene(self):
        config = get_config()
        path = config['data_dir']
        directory = path if path and op.isdir(path) else ""
        scene = self.model
        filename = scene.simple_name
        filename = re.sub(r'[\W]', '-', filename)
        fn = getSaveFileName(caption='Save scene to file',
                             filter='SVG (*.svg)',
                             suffix='svg',
                             selectFile=filename,
                             directory=directory,
                             parent=self)
        if not fn:
            return

        # Store old scene dialog path
        config['data_dir'] = op.dirname(fn)

        if not fn.endswith('.svg'):
            fn = f'{fn}.svg'

        with open(fn, 'w') as fout:
            fout.write(write_scene(scene))

    @Slot()
    def apply_editor_changes(self):
        self.scene_view.apply_editor_changes()

    @Slot()
    def decline_editor_changes(self):
        self.scene_view.decline_editor_changes()

    @Slot(bool)
    def toggle_snap_to_grid(self, is_checked):
        self.scene_view.snap_to_grid = is_checked

    @Slot()
    def resize_scene(self):
        dialog = ResizeSceneDialog(self.scene_view.size(), parent=self)
        move_to_cursor(dialog)
        if dialog.exec() != QDialog.Accepted:
            return

        self.scene_view.resize(dialog.scene_size)
        if not self.is_docked:
            width, height = self._compute_panel_size()
            self.resize(width, height)

    @Slot(bool)
    def design_mode_changed(self, is_checked):
        self.drawing_tool_bar.setVisible(is_checked)
        text = self.design_mode_text(is_checked)
        self.ac_design_mode.setToolTip(text)
        self.ac_design_mode.setStatusTip(text)
        self.ac_selection_tool.setChecked(is_checked)
        self.scene_view.design_mode = is_checked

    def design_mode_text(self, is_design_mode):
        if is_design_mode:
            return "Change to control mode"

        return "Change to design mode"

    # ----------------------------

    def setReadOnly(self, value):
        """This method is externally called for `AccessLevel` dependent view"""
        if not value and self.scene_view.design_mode:
            self.ac_design_mode.setChecked(False)
        self.ac_design_mode.setVisible(value)

    def setUnattached(self, editable: bool):
        """Set the scene panel to indicate an unattached scene"""
        self.ac_design_mode.toggled.disconnect()
        self.ac_design_mode.setIcon(icons.deviceClass)
        self.ac_design_mode.setCheckable(False)
        self.ac_design_mode.setToolTip("The scene is unattached.")
        self.ac_design_mode.setVisible(editable)

    def toggleAlwaysVisibleToolbar(self, value: bool) -> None:
        self.always_visible_tb.setVisible(value)

    def info(self):
        return create_scene_info(self)

    # ----------------------------
    # ToolBar and Action creators

    def create_toolbar_actions(self, connected_to_server):
        text = self.design_mode_text(self.scene_view.design_mode)
        self.ac_design_mode = QAction(icons.transform, text, self)
        self.ac_design_mode.setToolTip(text)
        self.ac_design_mode.setStatusTip(text)
        self.ac_design_mode.setCheckable(True)
        self.ac_design_mode.setChecked(self.scene_view.design_mode)
        self.ac_design_mode.setEnabled(connected_to_server)
        self.ac_design_mode.toggled.connect(self.design_mode_changed)

        self.ac_apply_all = QAction(icons.apply, 'Apply All Changes', self)
        self.ac_apply_all.triggered.connect(self.apply_editor_changes)
        self.ac_decline_all = QAction(icons.no, 'Decline All Changes', self)
        self.ac_decline_all.triggered.connect(self.decline_editor_changes)

        self.ac_save_scene = QAction(icons.saveAs, 'Save to file', self)
        self.ac_save_scene.triggered.connect(self.save_scene)
        self.qactions = []
        mode_qactions = self.create_mode_qactions()
        tool_qactions = [self._build_qaction(a)
                         for a in self.create_tool_actions()]
        link_qactions = [self._build_qaction(a)
                         for a in self.create_link_actions()]

        self._build_qaction_group(
            mode_qactions + tool_qactions + link_qactions)
        self.qactions.extend(mode_qactions)
        self.qactions.append(self._build_separator())

        self.qactions.extend(self.create_scene_tool_actions())
        self.qactions.append(self._build_separator())

        self.qactions.extend(tool_qactions)
        self.qactions.append(self._build_separator())

        link_menu = QMenu()
        for q_action in link_qactions:
            link_menu.addAction(q_action)
        link_action = QAction(icons.scenelink, "Link", self)
        link_action.setMenu(link_menu)
        self.qactions.extend([link_action, self._build_separator()])

        menu = QMenu()
        group_actions = self.create_group_tool_actions()
        for action in group_actions:
            q_action = self._build_qaction(action)
            menu.addAction(q_action)
        group_action = QAction(icons.group, "Group", self)
        group_action.setMenu(menu)
        self.qactions.extend([group_action, self._build_separator()])

        self.qactions.extend(self._build_qaction(a)
                             for a in self.create_clipboard_actions())

        self.qactions.append(self._build_separator())

        menu = QMenu()
        move_actions = self.create_move_actions()
        for action in move_actions:
            q_action = self._build_qaction(action)
            menu.addAction(q_action)
        move_action = QAction(icons.move, "Move", self)
        move_action.setMenu(menu)
        self.qactions.extend([move_action, self._build_separator()])

        menu = QMenu()
        align_actions = self.create_align_actions()
        for action in align_actions:
            q_action = self._build_qaction(action)
            menu.addAction(q_action)
        align_action = QAction(icons.alignLeft, "Align", self)
        align_action.setMenu(menu)
        self.qactions.extend([align_action, self._build_separator()])

        self.qactions.extend(self._build_qaction(a)
                             for a in self.create_order_actions())

    def create_mode_qactions(self):
        """ Create mode QActions and return list of them"""
        selection = CreateToolAction(
            tool_factory=SceneSelectionTool,
            icon=icons.cursorArrow,
            checkable=True,
            text="Selection Mode",
            tooltip="Select objects in the scene"
        )
        q_actions = [self._build_qaction(a) for a in [selection]]
        # Save a reference to the SceneSelectionTool QAction
        self.ac_selection_tool = q_actions[0]
        return q_actions

    def create_scene_tool_actions(self):
        show_grid_action = QAction(icons.grid, 'Snap to Grid', self)
        show_grid_action.setCheckable(True)
        show_grid_action.setChecked(self.scene_view.snap_to_grid)
        show_grid_action.triggered.connect(self.toggle_snap_to_grid)

        resize_scene_action = QAction(icons.resize, 'Resize Scene', self)
        resize_scene_action.triggered.connect(self.resize_scene)

        return [show_grid_action, resize_scene_action]

    def create_link_actions(self):
        """ Create link actions and return list of them"""
        actions = []
        actions.append(CreateToolAction(tool_factory=SceneLinkTool,
                                        icon=icons.scenelink,
                                        text="Add scene link",
                                        tooltip="Add scene link to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=WebLinkTool,
                                        icon=icons.weblink,
                                        text="Add web link",
                                        tooltip="Add web link to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=DeviceSceneLinkTool,
                                        icon=icons.deviceInstance,
                                        text="Add device scene link",
                                        tooltip="Add device link to scene",
                                        checkable=True))
        return actions

    def create_tool_actions(self):
        """ Create tool actions and return list of them"""
        actions = []
        actions.append(CreateToolAction(tool_factory=TextSceneTool,
                                        icon=icons.text, text="Add text",
                                        tooltip="Add text to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=LineSceneTool,
                                        icon=icons.line, text="Add line",
                                        tooltip="Add line to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=ArrowSceneTool,
                                        icon=icons.arrow,
                                        text="Add arrow",
                                        tooltip="Add arrow to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=RectangleSceneTool,
                                        icon=icons.rect, text="Add rectangle",
                                        tooltip="Add rectangle to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=StickerTool,
                                        icon=icons.sticker,
                                        text="Add a sticker widget",
                                        tooltip="Add sticker to scene",
                                        checkable=True))

        actions.append(CreateToolAction(tool_factory=PopupButtonTool,
                                        icon=icons.popup_sticker,
                                        text="Add a popup sticker widget",
                                        tooltip="Add a popup sticker widget "
                                                "to scene",
                                        checkable=True))

        actions.append(CreateToolAction(tool_factory=ImageRendererTool,
                                        icon=icons.image,
                                        text="Add an image to the scene",
                                        tooltip="Add image to scene",
                                        checkable=True))
        actions.append(CreateToolAction(tool_factory=InstanceStatusTool,
                                        icon=icons.statusOk,
                                        text="Add instance status",
                                        tooltip="Add instance status to scene",
                                        checkable=True))
        return actions

    def create_group_tool_actions(self):
        actions = []
        actions.append(GroupSceneAction(
            icon=icons.group,
            text="Group in Fixed Layout",
            shortcut=QKeySequence(Qt.CTRL + Qt.Key_G),
            tooltip="Group selected items"))
        actions.append(BoxVSceneAction(
            icon=icons.groupVertical,
            text="Group Vertically",
            tooltip="Group selected items in "
                    "vertical layout"))
        actions.append(BoxHSceneAction(
            icon=icons.groupHorizontal,
            text="Group Horizontally",
            tooltip="Group selected items in "
                    "horizontal layout"))
        actions.append(UngroupSceneAction(
            icon=icons.ungroup,
            shortcut=QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_G),
            text="Ungroup",
            tooltip="Ungroup selected items"))
        actions.append(GroupEntireSceneAction(
            icon=icons.entireWindow,
            text="Group Entire Window",
            tooltip="Group all items in "
                    "entire window"))
        return actions

    def create_clipboard_actions(self):
        actions = []
        actions.append(SceneSelectAllAction(icon=icons.selectAll,
                                            shortcut=QKeySequence.SelectAll,
                                            text="Select All",
                                            tooltip="Select all items of the "
                                                    "scene"))
        actions.append(SceneCutAction(icon=icons.editCut,
                                      shortcut=QKeySequence.Cut,
                                      text="Cut",
                                      tooltip="Cut selected items"))
        actions.append(SceneCopyAction(icon=icons.editCopy,
                                       shortcut=QKeySequence.Copy,
                                       text="Copy",
                                       tooltip="Copy selected items"))
        actions.append(ScenePasteAction(icon=icons.editPaste,
                                        shortcut=QKeySequence.Paste,
                                        text="Paste",
                                        tooltip="Paste selected items"))
        actions.append(ScenePasteReplaceAction(icon=icons.editPasteReplace,
                                               shortcut=QKeySequence.Replace,
                                               text="Paste and Replace",
                                               tooltip="Paste and replace the "
                                                       "associated device ID "
                                                       "of selected items"))
        actions.append(SceneDeleteAction(icon=icons.delete,
                                         shortcut=QKeySequence.Delete,
                                         text="Delete",
                                         tooltip="Delete selected items"))

        return actions

    def create_move_actions(self):
        actions = []

        actions.append(SceneMoveAction(
            icon=icons.arrowLeft,
            shortcut=QKeySequence.MoveToPreviousChar,
            text="Left",
            tooltip="Move item to the left"))
        actions.append(SceneMoveAction(
            icon=icons.arrowRight,
            shortcut=QKeySequence.MoveToNextChar,
            text="Right",
            tooltip="Move item to the right"))
        actions.append(SceneMoveAction(
            icon=icons.arrowUp,
            shortcut=QKeySequence.MoveToPreviousLine,
            text="Up",
            tooltip="Move item up"))
        actions.append(SceneMoveAction(
            icon=icons.arrowDown,
            shortcut=QKeySequence.MoveToNextLine,
            text="Down",
            tooltip="Move item down"))

        return actions

    def create_align_actions(self):
        actions = []

        actions.append(SceneAlignAction(
            icon=icons.alignLeft,
            text="Left",
            tooltip="Align items to left"))
        actions.append(SceneAlignAction(
            icon=icons.alignRight,
            text="Right",
            tooltip="Align items to right"))
        actions.append(SceneAlignAction(
            icon=icons.alignTop,
            text="Top",
            tooltip="Align items to top"))
        actions.append(SceneAlignAction(
            icon=icons.alignBottom,
            text="Bottom",
            tooltip="Align items to bottom"))

        return actions

    def create_order_actions(self):
        actions = []
        actions.append(SceneBringToFrontAction(icon=icons.bringToFront,
                                               text="Bring to Front",
                                               tooltip="Bring selected items "
                                                       "to front"))
        actions.append(SceneSendToBackAction(icon=icons.sendToBack,
                                             text="Send to Back",
                                             tooltip="Send selected items "
                                                     "to Back"))
        return actions

    def _build_qaction(self, sv_action):
        q_action = QAction(sv_action.icon, sv_action.text, self)
        q_action.setToolTip(sv_action.tooltip)
        q_action.setStatusTip(sv_action.tooltip)
        q_action.setCheckable(sv_action.checkable)
        q_action.triggered.connect(partial(sv_action.perform, self.scene_view))
        if sv_action.shortcut != QKeySequence.UnknownKey:
            if isinstance(sv_action.shortcut, QKeySequence):
                q_action.setShortcut(sv_action.shortcut)
            else:
                q_action.setShortcuts(
                    QKeySequence.keyBindings(sv_action.shortcut))
        return q_action

    def _build_qaction_group(self, actions):
        group = QActionGroup(self)
        for ac in actions:
            group.addAction(ac)
        actions[0].setChecked(True)  # Mark the first action

    def _build_separator(self):
        q_action = QAction(self)
        q_action.setSeparator(True)
        return q_action

    def _compute_panel_size(self):
        tb_height = 0
        if self.toolbar.isVisible():
            tb_height = self.toolbar.height()
        frame_width = self.model.width + QFRAME_PADDING
        frame_height = self.model.height + tb_height + QFRAME_PADDING
        return (frame_width, frame_height)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.model.simple_name}>"


class ResizableScrollArea(QScrollArea):
    """A QScrollArea which enables resizing of its child widget via mouse
    interaction
    """

    def __init__(self, child_widget, parent=None):
        super().__init__(parent=parent)

        self._resizing = False
        self._resize_type = ''
        self._child_widget = child_widget
        self._child_widget.installEventFilter(self)
        self.setMouseTracking(True)

    def eventFilter(self, obj, event):
        event_type = event.type()
        if event_type == QEvent.Enter:
            # Unset cursor when mouse is over child
            self.unsetCursor()
        elif event_type == QEvent.Leave:
            self._child_widget.mouse_left()
        return super().eventFilter(obj, event)

    def mouseMoveEvent(self, event):
        if not event.buttons():
            mouse_pos = event.pos()
            size = self._child_widget.size()
            dx = mouse_pos.x() - size.width()
            dy = mouse_pos.y() - size.height()
            right = 0 < dx < SCENE_BORDER_WIDTH
            bottom = 0 < dy < SCENE_BORDER_WIDTH
            if bottom and right:
                cursor = 'resize-diagonal-tlbr'
                self._resize_type = 'br'
            elif bottom and dx <= 0:
                cursor = 'resize-vertical'
                self._resize_type = 'b'
            elif right and dy <= 0:
                cursor = 'resize-horizontal'
                self._resize_type = 'r'
            else:
                cursor = 'arrow'
                self._resize_type = ''

            if not self._child_widget.design_mode:
                # Disallow unless the scene is in design mode!
                cursor = 'arrow'
                self._resize_type = ''

            self.setCursor(QT_CURSORS[cursor])
        elif self._resizing:
            mouse_pos = event.pos()
            size = QSize(self._child_widget.size())
            if "b" in self._resize_type:
                size.setHeight(mouse_pos.y())
            if "r" in self._resize_type:
                size.setWidth(mouse_pos.x())
            if (size.width() < SCENE_BORDER_WIDTH or
                    size.height() < SCENE_BORDER_WIDTH):
                return
            self._child_widget.resize(size)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self._resize_type != '':
            self._resizing = True
            event.accept()

    def mouseReleaseEvent(self, event):
        if self._resizing:
            self._resizing = False
            self._resize_type = ''
            self.unsetCursor()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Slot

from karabogui.graph.common.const import TF_SCALING, TF_TRANSLATION
from karabogui.graph.common.enums import ROITool
from karabogui.graph.common.roi.utils import ImageRegion

from .base import BaseROIController


class ImageROIController(BaseROIController):

    def __init__(self, plotItem):
        super().__init__(plotItem)
        plotItem.imageTransformed.connect(self._update_transform)

        # Enable movement wrt image pixels
        self.scaleSnap = True
        self.translateSnap = True

    def show(self, roi):
        super().show(roi)
        if self.plotItem.image_set:
            # Create a first ROI read
            self.update()

    def _show_roi_item(self, roi_item):
        super()._show_roi_item(roi_item)
        self._update_geometry(roi_item)

    def destroy(self):
        self.plotItem.imageTransformed.disconnect(self._update_transform)
        super().destroy()

    @Slot()
    def update(self):
        """Emits the ROI information, which can be either from the whole plot
           or the ROI rectangle."""
        if not self.plotItem.image_set or not self._updates_enabled:
            return

        image = self.plotItem.image

        # Get current ROI item
        if self._current_tool is ROITool.NoROI:
            current_item = None
        else:
            current_item = self._current_item[self._current_tool]

        # Get data of the region. If ROI is none, get the whole image.
        if current_item is None:
            region = ImageRegion(image, ImageRegion.Area,
                                 x_slice=slice(image.shape[1]),
                                 y_slice=slice(image.shape[0]))
        else:
            region = current_item.get_region(self.plotItem.imageItem)

        self.updated.emit(region)

    @Slot()
    def _update_geometry(self, roi=None):
        """Adjusts ROI item geometry with the current image transform"""
        if self._current_tool == ROITool.NoROI:
            return

        # If roi is None, transform all visible ROI items
        if roi is None:
            roi_items = self._rois[self._current_tool]
        else:
            roi_items = [roi]

        for item in roi_items:
            item.update_geometry_from_transform(
                self.plotItem.axes_transform[TF_SCALING],
                self.plotItem.axes_transform[TF_TRANSLATION], update=False)

    @Slot(object)
    def _set_current_item(self, roi_item, update=True):
        super()._set_current_item(roi_item, update)
        if update and self.plotItem.image_set:
            self.update()

    @Slot()
    def _update_transform(self):
        self._update_geometry()
        self.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QPoint, QSize, Qt, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QTableWidgetItem

from .utils import get_dialog_ui


class ReplaceDialog(QDialog):
    def __init__(self, devices, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('replacedialog.ui'), self)

        self.twTable.setRowCount(len(devices))
        for i, d in enumerate(devices):
            item = QTableWidgetItem(d)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.twTable.setItem(i, 0, item)

            item = QTableWidgetItem(d)
            self.twTable.setItem(i, 1, item)
            self.twTable.editItem(item)
        self.twTable.itemChanged.connect(self.onItemChanged)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)

    def mappedDevices(self):
        """
        A dict with the mapped devices is returned.
        """
        retval = {}
        for i in range(self.twTable.rowCount()):
            item_text = self.twTable.item(i, 1).text()
            retval[self.twTable.item(i, 0).text()] = item_text
        return retval

    @Slot(QTableWidgetItem)
    def onItemChanged(self, item):
        if item.column() != 1:
            return
        text = item.text()
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(len(text) > 0)


class SceneItemDialog(QDialog):
    """A dialog to modify the layout bounding rect coordinates
    """

    def __init__(self, x=0, y=0, title='SceneItem', max_x=1024, max_y=768,
                 parent=None):
        super().__init__(parent)
        filepath = get_dialog_ui('sceneitem_dialog.ui')
        uic.loadUi(filepath, self)
        self.setModal(False)
        # Fill the dialog with start values!
        self.ui_x.setValue(x)
        self.ui_x.setMaximum(max_x)
        self.ui_y.setValue(y)
        self.ui_y.setMaximum(max_y)
        self.setWindowTitle(title)
        if parent is not None:
            # place dialog accordingly!
            point = parent.rect().bottomRight()
            global_point = parent.mapToGlobal(point)
            self.move(global_point - QPoint(self.width(), 0))

    @property
    def x(self):
        return self.ui_x.value()

    @property
    def y(self):
        return self.ui_y.value()


class ResizeSceneDialog(QDialog):
    def __init__(self, size=QSize(), parent=None):
        super().__init__(parent)
        self.setModal(False)
        ui_path = get_dialog_ui('resizedialog.ui')
        uic.loadUi(ui_path, self)
        self.width_spinbox.setValue(size.width())
        self.height_spinbox.setValue(size.height())
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @property
    def scene_size(self):
        return QSize(self.width_spinbox.value(), self.height_spinbox.value())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import sys
from pathlib import Path

from qtpy.QtCore import QLine, QPoint, QRect
from qtpy.QtGui import QBrush, QImageReader, QPen, QPixmap, QPolygon
from qtpy.QtSvg import QSvgRenderer
from qtpy.QtWidgets import QDialog
from traits.api import Instance

from karabo.common.scenemodel.api import (
    ArrowPolygonModel, DeviceSceneLinkModel, ImageRendererModel,
    InstanceStatusModel, LineModel, PopupButtonModel, RectangleModel,
    SceneLinkModel, SceneTargetWindow, StickerModel, WebLinkModel,
    create_base64image)
from karabo.common.utils import get_arrowhead_points
from karabogui import messagebox
from karabogui.dialogs.api import (
    DeviceCapabilityDialog, SceneLinkDialog, TextDialog, TopologyDeviceDialog,
    WebDialog)
from karabogui.fonts import get_font_metrics
from karabogui.sceneview.bases import BaseSceneTool
from karabogui.sceneview.utils import calc_snap_pos
from karabogui.singletons.api import get_config
from karabogui.util import getOpenFileName


class TextSceneTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        dialog = TextDialog(alignment=True, parent=scene_view)
        if dialog.exec() == QDialog.Rejected:
            return

        model = dialog.label_model
        pos = event.pos()
        model.x = pos.x()
        model.y = pos.y()
        scene_view.add_models(model, initialize=True)
        scene_view.set_tool(None)


class LineSceneTool(BaseSceneTool):
    visible = True
    line = Instance(QLine)
    start_pos = Instance(QPoint)

    def draw(self, scene_view, painter):
        """Draw the line
        """
        if self.line is not None:
            painter.setPen(QPen())
            painter.drawLine(self.line)

    def mouse_down(self, scene_view, event):
        """A callback which is fired whenever the user clicks in the
        SceneView.
        """
        pos = event.pos()
        if scene_view.snap_to_grid:
            pos = calc_snap_pos(pos)
        self.start_pos = pos
        self.line = QLine(pos, pos)

    def mouse_move(self, scene_view, event):
        """A callback which is fired whenever the user moves the mouse
        in the SceneView.
        """
        if event.buttons() and self.line is not None:
            pos = event.pos()
            if scene_view.snap_to_grid:
                pos = calc_snap_pos(pos)
            self.line.setPoints(self.start_pos, pos)

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        if self.line is not None:
            model = LineModel(x1=self.line.x1(), y1=self.line.y1(),
                              x2=self.line.x2(), y2=self.line.y2(),
                              stroke="#000000")
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)
            scene_view.select_model(model)


class ArrowSceneTool(LineSceneTool):

    def draw(self, scene_view, painter):
        super().draw(scene_view, painter)
        if self.line is not None:
            p1 = self.line.p1()
            p2 = self.line.p2()
            hx1, hy1, hx2, hy2 = get_arrowhead_points(
                p1.x(), p1.y(), p2.x(), p2.y())
            hp1 = QPoint(hx1, hy1)
            hp2 = QPoint(hx2, hy2)

            # Paint arrowhead as a polygon of three points
            painter.setBrush(QBrush(painter.pen().color()))
            painter.drawPolygon(QPolygon([p2, hp1, hp2]))

    def mouse_up(self, scene_view, event):
        if self.line is not None:
            pos = event.pos()
            if scene_view.snap_to_grid:
                pos = calc_snap_pos(pos)
            x1 = self.start_pos.x()
            y1 = self.start_pos.y()
            x2 = pos.x()
            y2 = pos.y()
            hx1, hy1, hx2, hy2 = get_arrowhead_points(x1, y1, x2, y2)
            model = ArrowPolygonModel(
                x1=x1, y1=y1, x2=x2, y2=y2, hx1=hx1, hy1=hy1,
                hx2=hx2, hy2=hy2, stroke="#000000")
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)
            scene_view.select_model(model)


class RectangleSceneTool(BaseSceneTool):
    visible = True
    rect = Instance(QRect)
    start_pos = Instance(QPoint)

    def draw(self, scene_view, painter):
        """Draw the rect
        """
        if self.rect is not None:
            painter.setPen(QPen())
            painter.drawRect(self.rect)

    def mouse_down(self, scene_view, event):
        """A callback which is fired whenever the user clicks in the
        SceneView.
        """
        self.start_pos = event.pos()
        self.rect = QRect(self.start_pos, self.start_pos)

    def mouse_move(self, scene_view, event):
        """A callback which is fired whenever the user moves the mouse
        in the SceneView.
        """
        if event.buttons() and self.rect is not None:
            self.rect = QRect(self.start_pos, event.pos())

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        if self.rect is not None:
            model = RectangleModel(x=self.rect.x(), y=self.rect.y(),
                                   height=self.rect.height(),
                                   width=self.rect.width(),
                                   stroke="#000000")
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)


class SceneLinkTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        mouse_pos = event.pos()
        model = SceneLinkModel(x=mouse_pos.x(), y=mouse_pos.y(),
                               width=100, height=30)
        dialog = SceneLinkDialog(model, parent=scene_view)
        result = dialog.exec()
        if result == QDialog.Accepted:
            model.target = dialog.selectedScene
            model.target_window = dialog.selectedTargetWindow
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)


class DeviceSceneLinkTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        mouse_pos = event.pos()
        dialog = DeviceCapabilityDialog(parent=scene_view)
        result = dialog.exec()
        if result == QDialog.Accepted:
            device_id = dialog.device_id
            scene_name = dialog.capa_name

            _LINK_MARGIN = 10
            _LINK_SIZE_HIT = 30
            fm = get_font_metrics()
            model = DeviceSceneLinkModel(x=mouse_pos.x(), y=mouse_pos.y())
            width = max(fm.width(device_id) + _LINK_MARGIN, _LINK_SIZE_HIT)
            height = max(fm.height() + _LINK_MARGIN, _LINK_SIZE_HIT)
            model.width = width
            model.height = height

            # Link properties
            model.keys = [f"{device_id}.availableScenes"]
            model.text = f"{device_id}"
            model.target = scene_name
            model.target_window = SceneTargetWindow.Dialog
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)


class InstanceStatusTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        dialog = TopologyDeviceDialog(parent=scene_view)
        if dialog.exec() == QDialog.Rejected:
            return
        mouse_pos = event.pos()
        device_id = dialog.device_id
        model = InstanceStatusModel(
            x=mouse_pos.x(), y=mouse_pos.y(),
            width=30, height=30,
            keys=[f"{device_id}.deviceId"])
        scene_view.add_models(model, initialize=True)
        scene_view.set_tool(None)


class StickerTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        mouse_pos = event.pos()
        model = StickerModel(x=mouse_pos.x(), y=mouse_pos.y(),
                             width=100, height=100)
        scene_view.add_models(model, initialize=True)
        scene_view.set_tool(None)


class PopupButtonTool(StickerTool):
    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        mouse_pos = event.pos()
        model = PopupButtonModel(x=mouse_pos.x(), y=mouse_pos.y(),
                                 popup_width=160, popup_height=80)
        scene_view.add_models(model, initialize=True)
        scene_view.set_tool(None)


class WebLinkTool(BaseSceneTool):
    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        mouse_pos = event.pos()
        model = WebLinkModel(x=mouse_pos.x(), y=mouse_pos.y(),
                             width=100, height=30)
        dialog = WebDialog(parent=scene_view)
        if dialog.exec() == QDialog.Accepted:
            model.target = dialog.target
            scene_view.add_models(model, initialize=True)
            scene_view.set_tool(None)


class ImageRendererTool(BaseSceneTool):
    """This tool will place an `ImageRenderer` widget on the scene"""

    def mouse_down(self, scene_view, event):
        pass

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        fn = getOpenFileName(filter="Images (*.png *.jpg *.jpeg *.svg)",
                             directory=get_config()["data_dir"])
        if not fn:
            return

        get_config()["data_dir"] = str(Path(fn).parent)
        with open(fn, "rb") as fp:
            b = fp.read()

        limit = 102400  # 100 KB
        file_size = sys.getsizeof(b)
        if file_size > limit:
            scene_view.set_tool(None)
            messagebox.show_error("Selected image byte size is larger than "
                                  f"the allowed size of {limit // 1024}kB.",
                                  parent=scene_view)
            return

        image_format = str(QImageReader.imageFormat(fn), "utf-8")
        if image_format == "svg":
            renderer = QSvgRenderer(b)
            if not renderer.isValid() or renderer.defaultSize().isNull():
                scene_view.set_tool(None)
                messagebox.show_error(
                    "Selected image is not a valid svg file.",
                    parent=scene_view)
                return
            size = renderer.defaultSize()
        else:
            pixmap = QPixmap()
            pixmap.loadFromData(b)
            if pixmap.isNull():
                scene_view.set_tool(None)
                messagebox.show_error(
                    "Selected image is not a valid image file.",
                    parent=scene_view)
                return
            size = pixmap.size()

        mouse_pos = event.pos()
        image = create_base64image(image_format, b)
        model = ImageRendererModel(x=mouse_pos.x(), y=mouse_pos.y(),
                                   image=image,
                                   width=size.width(), height=size.height())
        scene_view.add_models(model, initialize=True)
        scene_view.set_tool(None)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Undefined

from karabogui.binding.api import FloatBinding

from ..alarm_dialog import LABEL_NOTE, AlarmDialog


def test_alarm_configuration_dialog(gui_app):
    binding = FloatBinding()
    config = {}
    dialog = AlarmDialog(binding, config)

    assert dialog.edit_alarmLow.text() == ""
    assert dialog.edit_warnLow.text() == ""
    assert dialog.edit_warnHigh.text() == ""
    assert dialog.edit_alarmHigh.text() == ""

    # Check the return values
    for v in dialog.values.values():
        assert v is Undefined

    model = {"alarmLow": 1.2}
    dialog = AlarmDialog(binding, model)
    assert dialog.edit_alarmLow.text() == "1.2"
    assert dialog.edit_warnLow.text() == ""
    assert dialog.edit_warnHigh.text() == ""
    assert dialog.edit_alarmHigh.text() == ""
    assert dialog.values["alarmLow"] == 1.2

    # Test a full model
    model = {"alarmLow": 1.0, "warnLow": 1.2,
             "warnHigh": 2.1, "alarmHigh": 5.7}
    dialog = AlarmDialog(binding, model)
    assert dialog.edit_alarmLow.text() == "1.0"
    assert dialog.edit_warnLow.text() == "1.2"
    assert dialog.edit_warnHigh.text() == "2.1"
    assert dialog.edit_alarmHigh.text() == "5.7"
    assert dialog.values["alarmLow"] == 1.0
    assert dialog.values["warnLow"] == 1.2
    assert dialog.values["warnHigh"] == 2.1
    assert dialog.values["alarmHigh"] == 5.7

    # Test an unordered model, this cannot happen on the field, because we
    # order!
    model = {"alarmLow": 1.2, "warnLow": 2.8,
             "warnHigh": 4.2, "alarmHigh": 0.8}
    dialog = AlarmDialog(binding, model)
    assert dialog.edit_alarmLow.text() == "1.2"
    assert dialog.edit_warnLow.text() == "2.8"
    assert dialog.edit_warnHigh.text() == "4.2"
    assert dialog.edit_alarmHigh.text() == "0.8"
    assert dialog.values["alarmLow"] == 0.8
    assert dialog.values["warnLow"] == 1.2
    assert dialog.values["warnHigh"] == 2.8
    assert dialog.values["alarmHigh"] == 4.2

    # Test unordered but with empty setting
    model = {"alarmLow": 1.2, "warnHigh": 4.2, "alarmHigh": 0.8}
    dialog = AlarmDialog(binding, model)
    assert dialog.edit_alarmLow.text() == "1.2"
    assert dialog.edit_warnLow.text() == ""
    assert dialog.edit_warnHigh.text() == "4.2"
    assert dialog.edit_alarmHigh.text() == "0.8"
    assert dialog.values["alarmLow"] == 0.8
    assert dialog.values["warnLow"] is Undefined
    assert dialog.values["warnHigh"] == 1.2
    assert dialog.values["alarmHigh"] == 4.2

    dialog.destroy()


def test_warning_note(gui_app):
    """ Test the warning message when the """
    binding = FloatBinding()
    error_message = "Please make sure the values are in the order:"
    model = {"alarmLow": 1.2, "warnLow": 2.8,
             "warnHigh": 4.2, "alarmHigh": Undefined}
    default_style = "QLabel#label_note {}"
    error_style = "QLabel#label_note {color:red;}"
    dialog = AlarmDialog(binding, model)
    dialog.edit_alarmHigh.setText("0.8")
    order = " < ".join(model.keys())
    text = f"{error_message} {order}"
    assert dialog.label_note.text() == text
    assert dialog.label_note.styleSheet() == error_style

    dialog.edit_alarmHigh.setText("8.2")
    assert dialog.label_note.text() == LABEL_NOTE
    assert dialog.label_note.styleSheet() == default_style

    # Keep a value empty
    dialog.edit_alarmHigh.setText("")
    assert dialog.label_note.text() == LABEL_NOTE
    assert dialog.label_note.styleSheet() == default_style

    # Include only non-empty field names in the warning.
    dialog.edit_warnHigh.setText("0.02")
    expected = f"{error_message} alarmLow < warnLow < warnHigh"
    assert dialog.label_note.text() == expected
    assert dialog.label_note.styleSheet() == error_style



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.widgets.hints import ElidingLabel


def test_elided_label(gui_app):
    label = ElidingLabel()
    label.setMaximumWidth(100)
    label.setText("Karabo")
    assert not label.isElided()
    label.setText("".join(15 * ["Karabo"]))
    assert label.isElided()
    label.setText("ThisIsKarabo")
    assert not label.isElided()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QDateTime, Qt

from karabogui.dialogs.api import ConfigurationFromPastDialog
from karabogui.events import KaraboEvent
from karabogui.singletons.mediator import Mediator
from karabogui.testing import GuiTestCase, singletons


class TestConfigurationFromPastDialog(GuiTestCase):

    def test_config_past_dialog(self):
        mediator = Mediator()
        network = mock.Mock()
        mediator.register_listener = mock.Mock()
        mediator.unregister_listener = mock.Mock()
        with singletons(mediator=mediator, network=network):
            dialog = ConfigurationFromPastDialog(instance_id="divvy")
            assert dialog.ui_instance_id.text() == "divvy"
            # Dialog registers to mediator
            mediator.register_listener.assert_called_once()

            # Clicking the show device requests navigation panel to show
            path = "karabogui.dialogs.configuration_from_past.broadcast_event"
            with mock.patch(path) as broadcast:
                self.click(dialog.ui_show_device)
                broadcast.assert_called_with(
                    KaraboEvent.ShowDevice, {"deviceId": "divvy",
                                             "showTopology": True})

            time = QDateTime.fromString("2021-07-15 09:07:42",
                                        "yyyy-MM-dd HH:mm:ss")
            dialog.ui_timepoint.setDateTime(time)
            self.click(dialog.ui_request)
            network.onGetConfigurationFromPast.assert_called_with(
                "divvy", preview=False, time=time.toUTC().toString(Qt.ISODate))

            # Click different time buttons
            for button in ("ui_one_week", "ui_one_day",
                           "ui_one_hour", "ui_ten_minutes"):
                prev_time = dialog.ui_timepoint.dateTime()
                self.click(getattr(dialog, button))
                self.process_qt_events()
                assert dialog.ui_timepoint.dateTime() != prev_time

            # Unregister and set successful
            dialog.done(1)
            mediator.unregister_listener.assert_called_once()

            # Accept dialog also requests
            dialog.ui_timepoint.setDateTime(time)
            dialog.accept()
            network.onGetConfigurationFromPast.assert_called_with(
                "divvy", preview=False, time=time.toUTC().toString(Qt.ISODate))

            # Network event and close
            dialog._event_network({"status": False})


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtCore import QPropertyAnimation, QRect, Qt, Signal, Slot
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLabel, QPushButton, QWidget

SWITCH_NORMAL = "NORMAL"
SWITCH_EXPERT = "EXPERT"

BUTTON_WIDTH = 22
FRAME_WIDTH = 50
FRAME_HEIGHT = 22
BUTTON_HEIGHT = 18
CENTER = (FRAME_HEIGHT - BUTTON_HEIGHT) // 2

BACKGROUND = "gray"
SWITCH_COLOR = "lightgreen"


class ClickFrame(QFrame):
    clicked = Signal()

    def mousePressEvent(self, event):
        self.clicked.emit()
        super().mousePressEvent(event)


class SwitchButton(QWidget):
    toggled = Signal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(5)

        self.switch_frame = ClickFrame(self)
        self.switch_frame.setFixedSize(FRAME_WIDTH, FRAME_HEIGHT)
        self.switch_frame.setStyleSheet(
            f"background-color: {BACKGROUND};"
            "border-radius: 9px;"
        )
        self.button = QPushButton(self.switch_frame)
        self.button.setFixedSize(BUTTON_WIDTH, BUTTON_HEIGHT)
        self.button.setStyleSheet(
            "background-color: white;"
            "border-radius: 9px;"
            f"border: 2px solid {BACKGROUND};"
        )
        self.button.move(2, CENTER)

        self.label = QLabel(SWITCH_NORMAL)
        self.label.setFixedSize(80, FRAME_HEIGHT)
        self.label.setStyleSheet(
            "font-size: 12px; "
            "font-weight: bold;")
        self.label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)

        self.animation = QPropertyAnimation(
            targetObject=self.button, propertyName=b"geometry")
        self.animation.setDuration(200)
        self.button.clicked.connect(self.toggle_switch)
        self.switch_frame.clicked.connect(self.toggle_switch)

        self.layout.addWidget(self.switch_frame)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)

        self.is_switched = False

    @Slot()
    def toggle_switch(self):
        self.animation.stop()
        if self.is_switched:
            self.animation.setStartValue(
                QRect(FRAME_WIDTH - BUTTON_WIDTH - 2, CENTER, BUTTON_WIDTH,
                      BUTTON_HEIGHT))
            self.animation.setEndValue(
                QRect(2, CENTER, BUTTON_WIDTH - 2, BUTTON_HEIGHT))
            self.switch_frame.setStyleSheet(
                f"background-color: {BACKGROUND};"
                "border-radius: 9px;")
            self.label.setText(SWITCH_NORMAL)
        else:
            self.animation.setStartValue(
                QRect(2, CENTER, BUTTON_WIDTH, BUTTON_HEIGHT))
            self.animation.setEndValue(
                QRect(FRAME_WIDTH - BUTTON_WIDTH - 2, CENTER, BUTTON_WIDTH - 2,
                      BUTTON_HEIGHT))
            self.switch_frame.setStyleSheet(
                f"background-color: {SWITCH_COLOR};"
                "border-radius: 9px;")
            self.label.setText(SWITCH_EXPERT)

        self.is_switched = not self.is_switched
        self.toggled.emit(self.is_switched)
        self.animation.start()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtGui import QValidator

from karabogui.testing import GuiTestCase

from ..object_handle import ObjectDuplicateDialog


class TestObjectDuplicateDialog(GuiTestCase):
    TEXT = 'You are about to create <b>{}</b> duplicate(s)'

    def setUp(self):
        super().setUp()
        simple_name = "foo"
        self.dialog = ObjectDuplicateDialog(simple_name)

    def test_default_values(self):
        start = 0
        end = 0
        assert self.dialog.sbStart.value() == start
        assert self.dialog.sbEnd.value() == end
        assert self.dialog.laText.text() == (
               self.TEXT.format(self._num_duplicates(start, end)))

    def test_set_higher_end_value(self):
        start = 0
        end = 10
        self.dialog.sbEnd.setValue(end)

        assert self.dialog.sbStart.value() == start
        assert self.dialog.sbEnd.value() == end
        assert self.dialog.laText.text() == \
               self.TEXT.format(self._num_duplicates(start, end))

    def test_set_higher_start_value(self):
        start = 10
        self.dialog.sbStart.setValue(start)

        assert self.dialog.sbStart.value() == start
        assert self.dialog.sbEnd.value() == start
        assert self.dialog.laText.text() == \
               self.TEXT.format(self._num_duplicates(start, start))

    def test_input_validity(self):
        self._assert_input("100", is_valid=True)
        self._assert_input("-1", is_valid=False)  # negative
        self._assert_input("1001", is_valid=False)  # exceeding
        self._assert_input("1,000", is_valid=False)  # comma-separated
        self._assert_input("1.00", is_valid=False)  # decimals

    def _assert_input(self, value, *, is_valid):
        valid = QValidator.Acceptable if is_valid else QValidator.Invalid
        pos = 0

        start_result, *_ = self.dialog.sbStart.validate(value, pos)
        assert start_result == valid
        end_result, *_ = self.dialog.sbEnd.validate(value, pos)
        assert end_result == valid

    def _num_duplicates(self, start, end):
        return (end - start) + 1



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.binding.builder import build_binding
from karabogui.binding.config import apply_configuration
from karabogui.binding.proxy import DeviceProxy, PropertyProxy
from karabogui.controllers.display.tests.image import (
    get_image_hash, get_pipeline_schema)
from karabogui.graph.common.const import AXIS_ITEMS

from ..webcam_graph import DisplayWebCamGraph


def test_webcam_graph_basics(gui_app):
    # setup
    schema = get_pipeline_schema()
    binding = build_binding(schema)
    root_proxy = DeviceProxy(binding=binding)

    output_proxy = PropertyProxy(root_proxy=root_proxy, path="output.data")
    img_proxy = PropertyProxy(root_proxy=root_proxy, path="output.data.image")
    controller = DisplayWebCamGraph(proxy=img_proxy)

    # test basics
    controller.create(None)
    assert controller.widget is not None

    image_hash = get_image_hash()
    apply_configuration(image_hash, output_proxy.binding)

    plotItem = controller._plot
    assert not plotItem.menuEnabled()

    image_shape = list(plotItem.imageItem.image.shape)
    assert image_shape == image_hash["image"]["dims"]

    # Assert axis are disabled
    for axis in AXIS_ITEMS:
        axis_item = plotItem.getAxis(axis)
        assert not axis_item.isVisible()
    assert not all(plotItem.vb.mouseEnabled())

    assert controller._plot.imageItem.image.shape == (30, 40)
    # Erase widget, a default with size 10, 10 is set
    controller.clear_widget()
    assert controller._plot.imageItem.image.shape == (10, 10)

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QRect
from qtpy.QtWidgets import QWidgetItem

from .shape import ShapeLayoutItem


class BaseLayout:
    """ This is a mix-in class intended to be inherited by all SceneView
    layouts. It implements generic parts of the QLayout virtual interface.
    """
    def __init__(self, model, *args):
        super().__init__(*args)
        self.model = model
        self._drawable_items = []

    def __iter__(self):
        """ Implement the Python iterator protocol for all layouts.
        """
        for i in range(self.count()):
            yield self.itemAt(i)

    def _add_layout(self, layout):
        """This MUST be called by inherited classes if they overload it!
        """
        self._drawable_items.append(layout)

    def _add_shape(self, shape):
        """ Use our special hacky workaround for adding shapes to the layout.
        """
        item = ShapeLayoutItem(shape)
        self._add_layout(item)
        self._drawable_items.append(item)

    def _add_widget(self, widget):
        """ Needs to be reimplemented in the inherited classes to add a widget.
        """
        raise NotImplementedError("BaseLayout._add_widget")

    def _remove_layout(self, layout):
        self.removeItem(layout)
        self._drawable_items.remove(layout)

    def _remove_shape(self, shape):
        for item in self:
            if isinstance(item, ShapeLayoutItem) and item.model is shape.model:
                # Call QLayout::removeItem()
                self.removeItem(item)
                self._drawable_items.remove(item)
                return

    def _remove_widget(self, widget):
        self.removeWidget(widget)

    def draw(self, painter):
        """Draw all the things which have a draw() method"""
        for item in self._drawable_items:
            item.draw(painter)

    def hide(self):
        for item in self:
            if isinstance(item, QWidgetItem):
                item.widget().hide()
            else:
                item.hide()

    def show(self):
        for item in self:
            if isinstance(item, QWidgetItem):
                item.widget().show()
            else:
                item.show()

    def set_geometry(self, rect):
        self.setGeometry(rect)

    def translate(self, offset):
        rect = QRect(self.model.x + offset.x(), self.model.y + offset.y(),
                     self.model.width, self.model.height)
        self.set_geometry(rect)

    # --------------------------------------------
    # QLayout Virtual Functions

    def geometry(self):
        return QRect(self.model.x, self.model.y, self.model.width,
                     self.model.height)

    def setGeometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.invalidate()  # Important! Force a full layout recalculation
        super().setGeometry(rect)

    def sizeHint(self):
        return self.geometry().size()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QDialog
from traits.api import Instance, Type, on_trait_change

from karabo.common.scenemodel.api import (
    DigitIconsModel, SelectionIconsModel, TextIconsModel)
from karabogui.binding.api import (
    BaseBinding, FloatBinding, IntBinding, StringBinding, get_binding_value)
from karabogui.controllers.api import (
    BaseBindingController, has_options, register_binding_controller)
from karabogui.controllers.icons_dialogs import (
    DigitDialog, IconItem, IconLabel, SelectionDialog, TextDialog)

NUMERICAL_BINDINGS = (FloatBinding, IntBinding)
MINIMUM_SIZE = (24, 24)


class _BaseIcons(BaseBindingController):

    current_item = Instance(IconItem)

    def create_widget(self, parent):
        widget = IconLabel(parent)
        action = QAction('Change Icons...', widget)
        action.triggered.connect(self._on_change_icons)
        widget.addAction(action)

        return widget

    def value_update(self, proxy):
        value = get_binding_value(proxy)
        if value is None:
            self.current_item = None
            return

        current_item = None
        for item in self.model.values:
            if self.condition(value, item):
                current_item = item
                break

        # Update the current item and invoke the pixmap changes
        self.current_item = current_item

    def condition(self, value, item):
        """Specifies the condition in which the widget will change the icon
           into.

           :param value: proxy value. Can be int, float, or string
           :param item: IconItem value from the collection self.model.values"""
        raise NotImplementedError("The subclass must specify the condition "
                                  "for the icon changes.")

    @on_trait_change('model.values')
    def _force_update(self):
        """Handle updates to the `IconData` collection for all subclasses.
        We make sure to convert from plain IconData => IconItem here.
        """
        def _state(item):
            NAMES = ['equal', 'value', 'image', 'data']
            return {k: getattr(item, k) for k in NAMES}

        converted = [IconItem(**_state(it))
                     if not isinstance(it, IconItem) else it
                     for it in self.model.values]
        self.model.trait_setq(values=converted)

        # Maybe update the widget
        if (self.widget is not None and
                get_binding_value(self.proxy) is not None):
            # Store a reference for the current item
            current_item = self.current_item

            self.value_update(self.proxy)

            # Check if current item is still the same and only the image
            # has changed, we set the pixmap manually
            if current_item is not None and self.current_item is current_item:
                self.widget.setPixmap(current_item.pixmap)

    def _on_change_icons(self):
        binding = self.proxy.binding
        dialog = self.dialog_klass(self.model.values, binding,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.values = dialog.items
            # The trait handler won't fire as are using a list, hence, we force
            # an update!
            self._force_update()

    def _current_item_changed(self, item):
        if item is None:
            self.widget.setDefaultPixmap()
        else:
            self.widget.setPixmap(item.pixmap)


@register_binding_controller(ui_name='Icons', binding_type=NUMERICAL_BINDINGS,
                             klassname='DigitIcons')
class DigitIcons(_BaseIcons):
    model = Instance(DigitIconsModel, args=())
    dialog_klass = Type(DigitDialog)

    def condition(self, value, item):
        """Check if the proxy value satisfy the comparison of one of
           the options."""
        return (value < float(item.value)
                or value == float(item.value)
                and item.equal)


@register_binding_controller(ui_name='Selection Icons',
                             klassname='SelectionIcons',
                             binding_type=BaseBinding,
                             is_compatible=has_options)
class SelectionIcons(_BaseIcons):
    model = Instance(SelectionIconsModel, args=())
    dialog_klass = Type(SelectionDialog)

    def binding_update(self, proxy):
        items = list(self.model.values)
        for opt in proxy.binding.options:
            if not any(opt == item.value for item in self.model.values):
                newItem = IconItem(value=opt)
                items.append(newItem)
        self.model.values = items

    def condition(self, value, item):
        """Check if the proxy value is the same as the string of one of the
           options"""
        return value == item.value


@register_binding_controller(ui_name='Icons', binding_type=StringBinding,
                             klassname='TextIcons')
class TextIcons(_BaseIcons):
    model = Instance(TextIconsModel, args=())
    dialog_klass = Type(TextDialog)

    def condition(self, value, item):
        """Check if the proxy value is the same as the string of one of the
           options"""
        return item.re.match(value)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Bool, Dict, HasStrictTraits, Instance, WeakRef

from karabo.common.api import walk_traits_object
from karabo.common.project.api import ProjectModel
from karabo.common.scenemodel.api import SceneModel, SceneTargetWindow
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.api import get_panel_wrangler

from .loading_watcher import ProjectLoadingWatcher

_reload_sessions = set()


def add_restore(project, scene_data):
    """Add a session for bookkeeping"""
    global _reload_sessions
    _reload_sessions = {s for s in _reload_sessions if not s.finished}
    active = ProjectRestoreSession(project=project, scene_data=scene_data)
    _reload_sessions.add(active)


class ProjectRestoreSession(HasStrictTraits):
    """An object which can restore the last session
    """
    # A project to be restored
    project = WeakRef(ProjectModel)
    # A boolean that is set once the scene view events are broadcasted
    finished = Bool(False)
    # Dictionary for scene model meta data {uuid: Dict of data}
    scene_data = Dict
    # The watcher running in background
    _watcher = Instance(ProjectLoadingWatcher)

    def _project_changed(self):
        if self.project is not None:
            self._watcher = ProjectLoadingWatcher(project=self.project)
            self._watcher.on_trait_change(self._finished, "finished_event")

    def _finished(self):
        """Traits notification handler to broadcast actions
        """
        scene_view = []

        def visitor(model):
            if isinstance(model, SceneModel) and model.uuid in self.scene_data:
                data = {"model": model}
                data.update(self.scene_data.pop(model.uuid))
                scene_view.append(data)

        walk_traits_object(self.project, visitor)

        # Sort according to index
        scene_view = sorted(scene_view, key=lambda scene: scene["index"])
        for data in scene_view:
            broadcast_event(KaraboEvent.RestoreSceneView, data)
        self._watcher.on_trait_change(self._finished, "finished_event",
                                      remove=True)
        self.finished = True


def get_restore_data():
    """Retrieve existing meta data for scenes to restore"""
    wrangler = get_panel_wrangler()
    scene_data = wrangler.scene_panel_data()

    data = {}
    for model, panel in scene_data.items():
        target_window = (SceneTargetWindow.MainWindow if panel.is_docked
                         else SceneTargetWindow.Dialog)
        position = panel.pos()
        x, y = position.x(), position.y()
        data[model.uuid] = {
            "target_window": target_window,
            "x": x,
            "y": y,
            "index": panel.index,
        }
    return data



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 9, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

import pytest

import karabo.common.scenemodel.tests as sm
from karabo.common.scenemodel.api import (
    LineModel, PathModel, RectangleModel, SceneModel)

from ..api import SceneView

DATA_DIR = op.join(op.abspath(op.dirname(sm.__file__)), 'data')
INKSCAPE_DIR = op.join(DATA_DIR, 'inkscape')


@pytest.fixture
def setup_view(gui_app):
    view = SceneView()
    view._set_scene_model(SceneModel())
    yield view


def test_line_shape(setup_view):
    view = setup_view
    assert len(view._scene_obj_cache) == 0
    model = LineModel(x1=0, y1=0, x2=1, y2=1)
    view.add_models(model)
    assert len(view._scene_obj_cache) == 1
    assert model in view._scene_obj_cache


def test_rect_shape(setup_view):
    view = setup_view
    assert len(view._scene_obj_cache) == 0
    model = RectangleModel(x=0, y=0, width=100, height=100)
    view.add_models(model)
    assert len(view._scene_obj_cache) == 1
    assert model in view._scene_obj_cache


def test_path_shape(setup_view):
    view = setup_view
    assert len(view._scene_obj_cache) == 0
    model = PathModel()
    model.svg_data = ("M 111.42857,389.50504 285.71429,546.6479 "
                      "428.57143,398.07647")
    view.add_models(model)
    assert len(view._scene_obj_cache) == 1
    assert model in view._scene_obj_cache



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import random
from io import StringIO
from pathlib import Path

from lxml import etree
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import (
    QCheckBox, QLabel, QPushButton, QVBoxLayout, QWizard, QWizardPage)

from karabogui import icons
from karabogui.singletons.api import get_config


def _get_notes():
    """Get the `Tips & Tricks` Notes and return a dictionary with
    `title` and `text` in a random fashion.
    """
    folder = Path(__file__).parent

    ret = {}

    def get_lines(fn):
        with open(fn) as fp:
            xml = fp.read()
            tree = etree.parse(StringIO(xml))
            title = tree.find('.//title').text
            ret.update({title: xml})

    files = [filename for filename in folder.rglob("*.html")]
    shuffle = random.sample(files, len(files))
    for fn in shuffle:
        get_lines(fn)

    return ret


LOGO_PATH = str(Path(icons.__file__).parent / Path('splash.png'))
LOGO_WIDTH = 100


class TipsTricksWizard(QWizard):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowTitle("Karabo Tips & Tricks")
        self.setWizardStyle(QWizard.ClassicStyle)
        checkbox = QCheckBox("Don't show tips at startup!")
        show_wizard = get_config()['wizard']
        checkbox.setChecked(not show_wizard)
        checkbox.stateChanged.connect(self.update_start)
        self.setButton(QWizard.CustomButton1, checkbox)
        self.setOption(QWizard.HaveCustomButton1)

        close_button = QPushButton("OK")
        close_button.clicked.connect(self.close)
        self.setButton(QWizard.CustomButton2, close_button)
        self.setOption(QWizard.HaveCustomButton2)

        layout = [QWizard.CustomButton1,
                  QWizard.BackButton,
                  QWizard.NextButton,
                  QWizard.CustomButton2]
        self.setButtonLayout(layout)

        self.button(QWizard.BackButton).setObjectName("BackButton")
        self.button(QWizard.NextButton).setObjectName("NextButton")

        self._build_information_pages()

    def sizeHint(self):
        return QSize(400, 200)

    def _build_information_pages(self):
        """Build basic information pages for the wizard"""
        pixmap = QPixmap(LOGO_PATH)
        pixmap = pixmap.scaledToWidth(LOGO_WIDTH)
        for title, text in _get_notes().items():
            page = QWizardPage(parent=self)
            page.setTitle(title)
            page.setPixmap(QWizard.WatermarkPixmap, pixmap)
            label = QLabel(text)
            label.setWordWrap(True)
            layout = QVBoxLayout()
            layout.addWidget(label)
            page.setLayout(layout)

            self.addPage(page)

    @Slot(int)
    def update_start(self, state):
        show_wizard = not bool(state)
        get_config()['wizard'] = show_wizard



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod

from traits.api import ABCHasStrictTraits, Dict, Property


class BaseAnalyzer(ABCHasStrictTraits):

    # A dictionary that contains stats. Override if `depends_on` is wanted
    stats = Property(Dict)

    @abstractmethod
    def _get_stats(self):
        """Reimplement to perform statistics specific to the analysis"""

    @abstractmethod
    def analyze(self, region, **config):
        """Reimplement to perform analysis in the input image region"""

    @abstractmethod
    def clear_data(self):
        """Reimplement to clear stored data based on stored schema"""



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt, QTimer, Slot
from qtpy.QtGui import QColor, QPainter, QPen, QPixmap
from qtpy.QtWidgets import QWidget

from karabogui.sceneview.const import GRID_STEP

GRID_COLOR = (200, 200, 200, 100)
RESIZE_THROTTLE = 100
PEN_SIZE = 1


class GridView(QWidget):
    """An object representing the grid view for a Karabo GUI scene.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._pen = QPen()
        self._pen.setWidth(PEN_SIZE)
        self._pen.setColor(QColor(*GRID_COLOR))

        # The update timer is required to throttle resize events!
        self._update_timer = QTimer(self)
        self._update_timer.setInterval(RESIZE_THROTTLE)
        self._update_timer.timeout.connect(self._needs_update)

        # By default the grid is not visible!
        self._visible = False
        self._pixmap = QPixmap()

    def _build_pixmap(self):
        """Build a pixmap drawing a grid layout"""
        if not self._visible:
            self._pixmap = QPixmap()
            return

        # Create the pixmap layer and fill transparent!
        pixmap = QPixmap(self.size())
        pixmap.fill(Qt.transparent)

        # Set our pen
        painter = QPainter(pixmap)
        painter.setPen(self._pen)
        # And draw the horizontal and vertical lines of the grid
        width = self.width()
        height = self.height()
        for y in range(0, height, GRID_STEP):
            painter.drawLine(0, y, width, y)
        for x in range(0, width, GRID_STEP):
            painter.drawLine(x, 0, x, height)

        self._pixmap = pixmap

    def resizeEvent(self, event):
        """Throttle the generation of the grid on resize events!"""
        self._update_timer.start()
        super().resizeEvent(event)

    @Slot()
    def _needs_update(self):
        self._update_timer.stop()
        self.showGrid(self._visible)

    def paintEvent(self, event):
        with QPainter(self) as painter:
            painter.drawPixmap(0, 0, self._pixmap)

    # Public interface
    # -----------------------------------------------------------------------

    def showGrid(self, value):
        """Public method to toggle and recompute the grid view"""
        self._visible = value
        self._build_pixmap()
        self.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QMenu
from traits.api import Instance, WeakRef

from karabo.common.scenemodel.api import WebCamGraphModel
from karabogui import icons
from karabogui.binding.api import ImageBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.graph.image.api import (
    KaraboImageNode, KaraboImagePlot, KaraboImageView, karabo_default_image)


@register_binding_controller(ui_name='WebCam Graph',
                             klassname='WebCamGraph',
                             binding_type=ImageBinding, priority=90,
                             can_show_nothing=False)
class DisplayWebCamGraph(BaseBindingController):
    model = Instance(WebCamGraphModel, args=())

    _plot = WeakRef(KaraboImagePlot)
    _image_node = Instance(KaraboImageNode, args=())

    def create_widget(self, parent):
        widget = KaraboImageView(parent=parent)
        widget.stateChanged.connect(self._change_model)

        self._plot = widget.plot()

        # Disable ViewBox Menu!
        self._plot.set_viewbox_menu(None)

        # Disable mouse and axis!
        self._plot.setMouseEnabled(False, False)
        self._plot.hide_all_axis()

        # Undock - Transient model data
        if not self.model.undock:
            menu = QMenu(widget)
            undock_action = QAction(icons.undock, "Undock", widget)
            undock_action.triggered.connect(self._undock_graph)
            menu.addAction(undock_action)
            self._plot.set_viewbox_menu(menu)

        # Colormap
        widget.add_colormap_action()
        widget.restore({"colormap": self.model.colormap})

        return widget

    def clear_widget(self):
        self._plot.setData(karabo_default_image())

    # -----------------------------------------------------------------------
    # Qt Slots

    def _undock_graph(self):
        """Undock the graph image but don't offer roi option"""
        model = self.model.clone_traits()
        model.undock = True
        broadcast_event(KaraboEvent.ShowUnattachedController,
                        {"model": model})

    def _change_model(self, content):
        self.model.trait_set(**content)

    # -----------------------------------------------------------------------

    def value_update(self, proxy):
        self._image_node.set_value(proxy.value)

        if not self._image_node.is_valid:
            return

        self._plot.setData(self._image_node.get_data())



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on September 22, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import (
    QItemSelection, QSortFilterProxyModel, QStringListModel, Qt, Slot)
from qtpy.QtWidgets import QAbstractItemView, QDialog, QDialogButtonBox

from karabo.common.api import Interfaces
from karabogui.topology.api import getTopology

from .utils import get_dialog_ui


class InstanceFilterModel(QSortFilterProxyModel):

    def __init__(self, topology, parent=None):
        super().__init__(parent)
        self.topology = topology
        self._interface = None
        self.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setFilterRole(Qt.DisplayRole)
        self.setFilterKeyColumn(0)

    def setInterface(self, text):
        """Set a interface bit for the filtering"""
        if text == "All Devices":
            self._interface = None
        else:
            self._interface = Interfaces[text]

    def filterAcceptsRow(self, source_row, source_parent):
        model = self.sourceModel()
        source_index = model.index(source_row, self.filterKeyColumn(),
                                   source_parent)
        if source_index.isValid():
            if self._interface is not None:
                deviceId = source_index.data()
                attrs = self.topology[deviceId, ...]
                interfaces = attrs.get("interfaces", 0)
                if not self._check_interface(interfaces, self._interface):
                    return False

        return super().filterAcceptsRow(source_row, source_parent)

    def _check_interface(self, mask, bit):
        return (mask & bit) == bit


class TopologyDeviceDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("topology_device_dialog.ui"), self)
        self.setModal(False)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.device_id = ""

        for interface in Interfaces:
            self.ui_interface.addItem(interface.name)

        topology = getTopology()["device"]
        self.filter_model = InstanceFilterModel(topology, parent=self)
        self.filter_model.setSourceModel(QStringListModel(topology))
        self.filter_model.setFilterFixedString("")

        self.ui_devices_view.setWordWrap(True)
        self.ui_devices_view.setAlternatingRowColors(True)
        self.ui_devices_view.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.ui_devices_view.setModel(self.filter_model)
        self.ui_filter.textChanged.connect(self.onFilterChanged)
        self.ui_clear.clicked.connect(self.clearFilter)
        self.ui_interface.currentIndexChanged.connect(self.onFilterChanged)

        selection = self.ui_devices_view.selectionModel()
        selection.selectionChanged.connect(self.onSelectionChanged)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)

    @Slot()
    def onFilterChanged(self):
        self.filter_model.setInterface(self.ui_interface.currentText())
        self.filter_model.setFilterFixedString(self.ui_filter.text())

    @Slot()
    def clearFilter(self):
        self.ui_interface.setCurrentIndex(0)
        self.ui_filter.setText("")
        self.onFilterChanged()

    @Slot(QItemSelection, QItemSelection)
    def onSelectionChanged(self, selected, unselected):
        enabled = len(selected) > 0
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def done(self, result):
        device_id = ""

        selection = self.ui_devices_view.selectionModel()
        index = selection.currentIndex()
        if index.isValid():
            index = self.filter_model.mapToSource(index)
            device_id = index.data()

        self.device_id = device_id
        super().done(result)
